"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from OpenGL.GL import *
import numpy as np


def set_uniform(shader, content, name):
    location = glGetUniformLocation(shader, name)
    if location == -1:
        raise ValueError(
            f"Uniform '{name}' not found in shader program {shader}.")

    if isinstance(content, int):
        glUniform1i(location, content)
    elif isinstance(content, float):
        glUniform1f(location, content)
    elif isinstance(content, np.ndarray):
        if content.ndim == 1:
            if content.shape[0] == 2:
                glUniform2f(location, *content)
            elif content.shape[0] == 3:
                glUniform3f(location, *content)
            else:
                raise ValueError(
                    f"Unsupported 1D array size: {content.shape}.")
        elif content.ndim == 2:
            if content.shape == (4, 4):
                glUniformMatrix4fv(location, 1, GL_FALSE,
                                   content.T.astype(np.float32))
            else:
                raise ValueError(
                    f"Unsupported 2D array size: {content.shape}.")
        else:
            raise ValueError(f"Unsupported array dimension: {content.ndim}.")
    else:
        raise TypeError(
            f"Unsupported type for uniform '{name}': {type(content)}.")



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from OpenGL.GL import *
from math import radians, tan
import numpy as np
from PySide6 import QtCore, QtGui, QtOpenGLWidgets
from q3dviewer.utils.maths import frustum, euler_to_matrix, makeT


class BaseGLWidget(QtOpenGLWidgets.QOpenGLWidget):
    def __init__(self, parent=None):
        QtOpenGLWidgets.QOpenGLWidget.__init__(self, parent)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)
        self.reset()
        self._fov = 60
        self.items = []
        self.keyTimer = QtCore.QTimer()
        self.color = np.array([0, 0, 0, 0])
        self.dist = 40
        self.euler = np.array([np.pi/3, 0, np.pi/4])
        self.center = np.array([0, 0, 0.])
        self.active_keys = set()
        self.show_center = False
        self.enable_show_center = True
        self.view_need_update = True
        self.view_matrix = self.get_view_matrix()
        self.projection_matrix = self.get_projection_matrix()

    def keyPressEvent(self, ev: QtGui.QKeyEvent):
        if ev.key() == QtCore.Qt.Key_Up or  \
            ev.key() == QtCore.Qt.Key_Down or \
            ev.key() == QtCore.Qt.Key_Left or \
            ev.key() == QtCore.Qt.Key_Right or \
            ev.key() == QtCore.Qt.Key_Z or \
            ev.key() == QtCore.Qt.Key_X or \
            ev.key() == QtCore.Qt.Key_A or \
            ev.key() == QtCore.Qt.Key_D or \
            ev.key() == QtCore.Qt.Key_W or \
            ev.key() == QtCore.Qt.Key_S:
            self.active_keys.add(ev.key())
        self.active_keys.add(ev.key())

    def keyReleaseEvent(self, ev: QtGui.QKeyEvent):
        self.active_keys.discard(ev.key())

    def current_width(self):
        """
        Return the current width of the widget.
        """
        return int(self.width() * self.devicePixelRatioF())

    def current_height(self):
        """
        Return the current height of the widget.
        """
        return int(self.height() * self.devicePixelRatioF())

    def reset(self):
        pass

    def add_item(self, item):
        """
        Add the item to the glwidget.
        """
        self.items.append(item)
        item.set_glwidget(self)
        
    def remove_item(self, item):
        """
        Remove the item from the glwidget.
        """
        self.items.remove(item)
        item.set_glwidget(None)

    def clear(self):
        """
        Remove all items from the glwidget.
        """
        for item in self.items:
            item.set_glwidget(None)
        self.items = []
        
    def initializeGL(self):
        """
        the method is herted from QOpenGLWidget, 
        and it is called when the widget is first shown.
        """
        for item in self.items:
            item.initialize()
        # initialize the projection matrix and model view matrix
        self.projection_matrix = self.get_projection_matrix()
        self.update_model_projection()
        self.view_matrix = self.get_view_matrix()
        self.update_model_view()

    def set_view_matrix(self, view_matrix):
        self.view_matrix = view_matrix
        self.view_need_update = False

    def mouseReleaseEvent(self, ev):
        if hasattr(self, 'mousePos'):
            delattr(self, 'mousePos')

    def set_dist(self, dist):
        self.dist = dist
        self.view_need_update = True

    def update_dist(self, delta):
        self.dist += delta
        if self.dist < 0.1:
            self.dist = 0.1
        self.view_need_update = True

    def wheelEvent(self, ev):
        delta = ev.angleDelta().x()
        if delta == 0:
            delta = ev.angleDelta().y()
        self.update_dist(-delta * self.dist * 0.001)
        self.show_center = True

    def mouseMoveEvent(self, ev):
        lpos = ev.localPos()
        if not hasattr(self, 'mousePos'):
            self.mousePos = lpos
        diff = lpos - self.mousePos
        self.mousePos = lpos
        if ev.buttons() == QtCore.Qt.MouseButton.RightButton:
            rot_speed = 0.2
            dyaw = radians(-diff.x() * rot_speed)
            droll = radians(-diff.y() * rot_speed)
            self.rotate(droll, 0, dyaw)
        elif ev.buttons() == QtCore.Qt.MouseButton.LeftButton:
            Rwc = euler_to_matrix(self.euler)
            Kinv = np.linalg.inv(self.get_K())
            dist = max(self.dist, 0.5)
            self.center += Rwc @ Kinv @ np.array([-diff.x(), diff.y(), 0]) * dist
        self.show_center = True
        self.view_need_update = True

    def set_center(self, center):
        self.center = center
        self.view_need_update = True

    def paintGL(self):
        # if the camera is moved, update the model view matrix.
        if self.view_need_update:
            self.view_matrix = self.get_view_matrix()
            self.view_need_update = False
        self.update_model_view()

        # set the background color
        bgcolor = self.color
        glClearColor(*bgcolor)
        glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT)
        for item in self.items:
            if not item.visible():
                continue
            if not item.is_initialized():
                """
                The item may not be initialized if it is added
                after the widget is shown, so we need to initialize it here.
                """
                item.initialize()
            glMatrixMode(GL_MODELVIEW)
            glPushMatrix()
            glPushAttrib(GL_ALL_ATTRIB_BITS)
            try:
                item.paint()
            finally:
                glPopAttrib()
                glMatrixMode(GL_MODELVIEW)
                glPopMatrix()
        
        # Show center as a point if updated by mouse move event
        if self.enable_show_center and self.show_center:
            point_size = np.clip((self.get_K()[0, 0] / self.dist), 10, 100)
            glPointSize(point_size)
            glBegin(GL_POINTS)
            glColor3f(1.0, 0.0, 0.0)  # Red color for the center point
            glVertex3f(*self.center)
            glEnd()
            self.show_center = False
    
    def update_movement(self):
        """
        Update the movement of the camera based on the active keys.
        """
        if not self.active_keys:
            return
        rot_speed = 0.5
        trans_speed = max(self.dist * 0.005, 0.1)
        # Handle rotation keys
        if QtCore.Qt.Key_Up in self.active_keys:
            self.rotate(radians(rot_speed), 0, 0)
            self.view_need_update = True
        if QtCore.Qt.Key_Down in self.active_keys:
            self.rotate(radians(-rot_speed), 0, 0)
            self.view_need_update = True
        if QtCore.Qt.Key_Left in self.active_keys:
            self.rotate(0, 0, radians(rot_speed))
            self.view_need_update = True
        if QtCore.Qt.Key_Right in self.active_keys:
            self.rotate(0, 0, radians(-rot_speed))
            self.view_need_update = True
        # Handle zoom keys
        xz_keys = {QtCore.Qt.Key_Z, QtCore.Qt.Key_X}
        if self.active_keys & xz_keys:
            Rwc = euler_to_matrix(self.euler)
            if QtCore.Qt.Key_Z in self.active_keys:
                self.center += Rwc @ np.array([0, 0, -trans_speed])
                self.view_need_update = True
            if QtCore.Qt.Key_X in self.active_keys:
                self.center += Rwc @ np.array([0, 0, trans_speed])
                self.view_need_update = True
        # Handle translation keys on the z plane
        dir_keys = {QtCore.Qt.Key_W, QtCore.Qt.Key_S, QtCore.Qt.Key_A, QtCore.Qt.Key_D}
        if self.active_keys & dir_keys:
            Rz = euler_to_matrix([0, 0, self.euler[2]])
            if QtCore.Qt.Key_W in self.active_keys:
                self.center += Rz @ np.array([0, trans_speed, 0])
                self.view_need_update = True
            if QtCore.Qt.Key_S in self.active_keys:
                self.center += Rz @ np.array([0, -trans_speed, 0])
                self.view_need_update = True
            if QtCore.Qt.Key_A in self.active_keys:
                self.center += Rz @ np.array([-trans_speed, 0, 0])
                self.view_need_update = True
            if QtCore.Qt.Key_D in self.active_keys:
                self.center += Rz @ np.array([trans_speed, 0, 0])
                self.view_need_update = True

    def update_model_view(self):
        glMatrixMode(GL_MODELVIEW)
        glLoadMatrixf(self.view_matrix.T)
        
    def get_view_matrix(self):
        two = self.center # the origin(center) in the world frame
        tco = np.array([0, 0, self.dist]) # the origin(center) in camera frame
        Rwc = euler_to_matrix(self.euler)
        twc = two + Rwc @ tco
        Rcw = Rwc.T
        tcw = -Rcw @ twc
        Tcw = makeT(Rcw, tcw)
        return Tcw

    def set_cam_position(self, **kwargs):
        center = kwargs.get('center', None)
        distance = kwargs.get('distance', None)
        euler = kwargs.get('euler', None)
        if center is not None:
            self.set_center(center)
        if distance is not None:
            self.set_dist(distance)
        if euler is not None:
            self.set_euler(euler)
    
    def set_euler(self, euler):
        self.euler = euler
        self.view_need_update = True

    def set_color(self, color):
        self.color = color

    def update(self):
        self.update_movement()
        super().update()

    def update_model_projection(self):
        glMatrixMode(GL_PROJECTION)
        glLoadMatrixf(self.projection_matrix.T)

    def get_projection_matrix(self):
        w, h = self.current_width(), self.current_height()
        dist = self.dist
        near = dist * 0.001
        far = dist * 10000.
        r = near * tan(0.5 * radians(self._fov))
        t = r * h / w
        matrix = frustum(-r, r, -t, t, near, far)
        return matrix

    def get_K(self):
        project_matrix = self.get_projection_matrix()
        width = self.current_width()
        height = self.current_height()
        fx = project_matrix[0, 0] * width / 2
        fy = project_matrix[1, 1] * height / 2
        cx = width / 2
        cy = height / 2
        K = np.array([
            [fx, 0, cx],
            [0, fy, cy],
            [0, 0, 1]
        ])
        return K
            
    def rotate(self, rx=0, ry=0, rz=0):
        # update the euler angles
        self.euler += np.array([rx, ry, rz])
        self.euler[2] = (self.euler[2] + np.pi) % (2 * np.pi) - np.pi
        self.euler[1] = (self.euler[1] + np.pi) % (2 * np.pi) - np.pi
        self.euler[0] = np.clip(self.euler[0], 0, np.pi)

    def change_show_center(self, state):
        self.enable_show_center = state

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.projection_matrix = self.get_projection_matrix()
        self.update_model_projection()

    def capture_frame(self):
        self.makeCurrent()  # Ensure the OpenGL context is current
        width = self.current_width()
        height = self.current_height()
        pixels = glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE)
        frame = np.frombuffer(pixels, dtype=np.uint8).reshape(height, width, 3)
        frame = np.flip(frame, 0)
        return frame



#!/usr/bin/env python3

"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.utils.maths import make_transform
import q3dviewer as q3d
import rospy
from nav_msgs.msg import Odometry
from sensor_msgs.msg import PointCloud2
import numpy as np
import random
from sensor_msgs.msg import Image
from q3dviewer.utils.convert_ros_msg import convert_pointcloud2_msg, convert_odometry_msg, convert_image_msg

viewer = None
point_num_per_scan = None
color_mode = None
auto_set_color_mode = True

def odom_cb(data):
    global viewer
    transform, _ = convert_odometry_msg(data)
    viewer['odom'].set_transform(transform)


def scan_cb(data):
    global viewer
    global point_num_per_scan
    global auto_set_color_mode
    cloud, fields, _ = convert_pointcloud2_msg(data)
    if (cloud.shape[0] > point_num_per_scan):
        idx = random.sample(range(cloud.shape[0]), point_num_per_scan)
        cloud = cloud[idx]
    if 'rgb' in fields and auto_set_color_mode:
        print("Set color mode to RGB")
        viewer['map'].set_color_mode('RGB')
        auto_set_color_mode = False
    viewer['map'].set_data(data=cloud, append=True)
    viewer['scan'].set_data(data=cloud)


def image_cb(data):
    image, _ = convert_image_msg(data)
    viewer['img'].set_data(data=image)


def main():
    rospy.init_node('ros_viewer', anonymous=True)

    global viewer
    global point_num_per_scan
    point_num_per_scan = 10000
    map_item = q3d.CloudIOItem(size=1, alpha=0.1, color_mode='I')
    scan_item = q3d.CloudItem(
        size=2, alpha=1, color_mode='FLAT', color='#ffffff')
    odom_item = q3d.AxisItem(size=0.5, width=5)
    grid_item = q3d.GridItem(size=1000, spacing=20)
    img_item = q3d.ImageItem(pos=np.array([0, 0]), size=np.array([800, 600]))

    app = q3d.QApplication(['ROS Viewer'])
    viewer = q3d.Viewer(name='ROS Viewer')

    viewer.add_items({'map': map_item, 'scan': scan_item,
                     'odom': odom_item, 'grid': grid_item,
                     'img': img_item})

    point_num_per_scan = rospy.get_param("scan_num", 100000)
    print("point_num_per_scan: %d" % point_num_per_scan)
    rospy.Subscriber(
        "/cloud_registered", PointCloud2, scan_cb,
        queue_size=1, buff_size=2**24)
    rospy.Subscriber(
        "/odometry", Odometry, odom_cb, queue_size=1, buff_size=2**24)
    rospy.Subscriber('/image', Image, image_cb)

    viewer.show()
    app.exec()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
import q3dviewer as q3d
from PySide6.QtWidgets import QVBoxLayout, QListWidget, QListWidgetItem, QPushButton, QDoubleSpinBox, QCheckBox, QLineEdit, QMessageBox, QLabel, QHBoxLayout, QDockWidget, QWidget, QComboBox
from PySide6.QtCore import QTimer
from q3dviewer.tools.cloud_viewer import ProgressDialog, FileLoaderThread
from PySide6 import QtCore
from PySide6.QtGui import QKeyEvent
from q3dviewer import GLWidget
import imageio.v2 as imageio
import os


def recover_center_euler(Twc, dist):
    Rwc = Twc[:3, :3]  # Extract rotation
    twc = Twc[:3, 3]   # Extract translation
    tco = np.array([0, 0, dist])  # Camera frame origin
    two = twc - Rwc @ tco  # Compute center
    euler = q3d.matrix_to_euler(Rwc)
    return two, euler


class KeyFrame:
    def __init__(self, Twc, lin_vel=10, ang_vel=np.pi/3, stop_time=0):
        self.Twc = Twc
        self.lin_vel = lin_vel
        self.ang_vel = ang_vel  # rad/s
        self.stop_time = stop_time
        self.item = q3d.FrameItem(Twc, width=3, color='#0000FF')


class CustomGLWidget(GLWidget):
    def __init__(self, viewer):
        super().__init__()
        self.viewer = viewer  # Add a viewer handle

    def keyPressEvent(self, ev: QKeyEvent):
        if ev.key() == QtCore.Qt.Key_Space:
            self.viewer.add_key_frame()
        elif ev.key() == QtCore.Qt.Key_Delete:
            self.viewer.del_key_frame()
        elif ev.key() == QtCore.Qt.Key_C:
            self.viewer.dock.show()
        super().keyPressEvent(ev)

class CMMViewer(q3d.Viewer):
    """
    This class is a subclass of Viewer, which is used to create a cloud movie maker.
    """
    def __init__(self, **kwargs):
        self.key_frames = []
        self.video_path = os.path.join(os.path.expanduser("~"), "output.mp4")
        super().__init__(**kwargs, gl_widget_class=lambda: CustomGLWidget(self))
        # for drop cloud file
        self.setAcceptDrops(True)

    def add_control_panel(self, main_layout):
        """
        Add a control panel to the viewer.
        """
        # Create a vertical layout for the settings
        setting_layout = QVBoxLayout()

        # Buttons to add and delete key frames
        add_button = QPushButton("Add Key Frame (Key Space)")
        add_button.clicked.connect(self.add_key_frame)
        setting_layout.addWidget(add_button)
        del_button = QPushButton("Delete Key Frame (Key Delete)")
        del_button.clicked.connect(self.del_key_frame)
        setting_layout.addWidget(del_button)

        # Add play/stop button
        self.play_button = QPushButton("Play")
        self.play_button.clicked.connect(self.toggle_playback)
        setting_layout.addWidget(self.play_button)

        # add a timer to play the frames
        self.timer = QTimer()
        self.timer.timeout.connect(self.play_frames)
        self.current_frame_index = 0
        self.is_playing = False
        self.is_recording = False

        # Add record checkbox
        self.record_checkbox = QCheckBox("Record")
        self.record_checkbox.stateChanged.connect(self.toggle_recording)
        setting_layout.addWidget(self.record_checkbox)

        # Add video path setting
        video_path_layout = QHBoxLayout()
        label_video_path = QLabel("Video Path:")
        video_path_layout.addWidget(label_video_path)
        self.video_path_edit = QLineEdit()
        self.video_path_edit.setText(self.video_path)
        self.video_path_edit.textChanged.connect(self.update_video_path)
        video_path_layout.addWidget(self.video_path_edit)
        setting_layout.addLayout(video_path_layout)

        # Add codec setting
        codec_layout = QHBoxLayout()
        label_codec = QLabel("Codec:")
        codec_layout.addWidget(label_codec)
        self.codec_combo = QComboBox()
        self.codec_combo.addItems(["mjpeg", "mpeg4", "libx264", "libx265"])
        codec_layout.addWidget(self.codec_combo)
        setting_layout.addLayout(codec_layout)

        # Add a list of key frames
        self.frame_list = QListWidget()
        setting_layout.addWidget(self.frame_list)
        self.frame_list.itemSelectionChanged.connect(self.on_select_frame)
        self.frame_list.itemDoubleClicked.connect(self.on_double_click_frame)
        self.installEventFilter(self)

        # Add spin boxes for linear / angular velocity and stop time
        self.lin_vel_spinbox = QDoubleSpinBox()
        self.lin_vel_spinbox.setPrefix("Linear Velocity (m/s): ")
        self.lin_vel_spinbox.setRange(0, 1000)
        self.lin_vel_spinbox.valueChanged.connect(self.set_frame_lin_vel)
        setting_layout.addWidget(self.lin_vel_spinbox)

        self.lin_ang_spinbox = QDoubleSpinBox()
        self.lin_ang_spinbox.setPrefix("Angular Velocity (deg/s): ")
        self.lin_ang_spinbox.setRange(0, 360)
        self.lin_ang_spinbox.valueChanged.connect(self.set_frame_ang_vel)
        setting_layout.addWidget(self.lin_ang_spinbox)

        self.stop_time_spinbox = QDoubleSpinBox()
        self.stop_time_spinbox.setPrefix("Stop Time: ")
        self.stop_time_spinbox.setRange(0, 100)
        self.stop_time_spinbox.valueChanged.connect(self.set_frame_stop_time)
        setting_layout.addWidget(self.stop_time_spinbox)
        
        setting_layout.setAlignment(QtCore.Qt.AlignTop)

        # Create a dock widget for the settings
        dock_widget = QDockWidget("Settings", self)
        dock_widget.setWidget(QWidget())
        dock_widget.widget().setLayout(setting_layout)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock_widget)

        # Add the dock widget to the main layout
        main_layout.addWidget(dock_widget)
        self.dock = dock_widget


    def update_video_path(self, path):
        self.video_path = path

    def add_key_frame(self):
        view_matrix = self.glwidget.view_matrix
        # Get camera pose in world frame
        Twc = np.linalg.inv(view_matrix)
        if self.key_frames:
            prev = self.key_frames[-1]
            key_frame = KeyFrame(Twc,
                                 lin_vel=prev.lin_vel, 
                                 ang_vel=prev.ang_vel,
                                 stop_time=prev.stop_time)
        else:
            key_frame = KeyFrame(Twc)
        self.key_frames.append(key_frame)
        # visualize this key frame using FrameItem
        self.glwidget.add_item(key_frame.item)
        # move the camera back to 0.5 meter, let the user see the frame
        # self.glwidget.update_dist(0.5)
        # Add the key frame to the Qt ListWidget
        item = QListWidgetItem(f"Frame {len(self.key_frames)}")
        self.frame_list.addItem(item)
        self.frame_list.setCurrentRow(len(self.key_frames) - 1)

    def del_key_frame(self):
        current_index = self.frame_list.currentRow()
        if current_index < 0:
            return
        self.glwidget.remove_item(self.key_frames[current_index].item)
        self.key_frames.pop(current_index)
        self.frame_list.itemSelectionChanged.disconnect(self.on_select_frame)
        self.frame_list.takeItem(current_index)
        self.frame_list.itemSelectionChanged.connect(self.on_select_frame)
        self.on_select_frame()
        # Update frame labels
        for i in range(len(self.key_frames)):
            self.frame_list.item(i).setText(f"Frame {i + 1}")
    
    def on_select_frame(self):
        current = self.frame_list.currentRow()
        if current < 0:
            return
        for i, frame in enumerate(self.key_frames):
            if i == current:
                # Highlight the selected frame
                frame.item.set_color('#FF0000')
                frame.item.set_line_width(5)
                # show current frame's parameters in the spinboxes
                self.lin_vel_spinbox.setValue(frame.lin_vel)
                self.lin_ang_spinbox.setValue(np.rad2deg(frame.ang_vel))
                self.stop_time_spinbox.setValue(frame.stop_time)
            else:
                frame.item.set_color('#0000FF')
                frame.item.set_line_width(3)

    def set_frame_lin_vel(self, value):
        current_index = self.frame_list.currentRow()
        if current_index < 0:
            return
        self.key_frames[current_index].lin_vel = value

    def set_frame_ang_vel(self, value):
        current_index = self.frame_list.currentRow()
        if current_index < 0:
            return
        self.key_frames[current_index].ang_vel = np.deg2rad(value)

    def set_frame_stop_time(self, value):
        current_index = self.frame_list.currentRow()
        if current_index < 0:
            return
        self.key_frames[current_index].stop_time = value

    def on_double_click_frame(self, item):
        current_index = self.frame_list.row(item)
        if current_index < 0:
            return
        Twc = self.key_frames[current_index].Twc
        center, euler = recover_center_euler(Twc, self.glwidget.dist)
        self.glwidget.set_cam_position(center=center,
                                       euler=euler)


    def create_frames(self):
        """
        Create the frames for playback by interpolating between key frames.
        """
        self.frames = []
        dt = 1 / float(self.update_interval)
        for i in range(len(self.key_frames) - 1):
            current_frame = self.key_frames[i]
            if current_frame.stop_time > 0:
                num_steps = int(current_frame.stop_time / dt)
                for j in range(num_steps):
                    self.frames.append([i, current_frame.Twc])
            next_frame = self.key_frames[i + 1]
            Ts = q3d.interpolate_pose(current_frame.Twc, next_frame.Twc,
                                      current_frame.lin_vel,
                                      current_frame.ang_vel,
                                      dt)
            for T in Ts:
                self.frames.append([i, T])
        
        print(f"Total frames: {len(self.frames)}")
        print(f"Total time: {len(self.frames) * dt:.2f} seconds")

    def toggle_playback(self):
        if self.is_playing:
            self.stop_playback()
        else:
            self.start_playback()

    def start_playback(self):
        if self.key_frames:
            self.create_frames()
            self.current_frame_index = 0
            self.timer.start(self.update_interval)  # Adjust the interval as needed
            self.is_playing = True
            self.play_button.setStyleSheet("background-color: red")
            self.play_button.setText("Playing")
            self.record_checkbox.setEnabled(False)
            if self.is_recording is True:
                self.start_recording()

    def stop_playback(self):
        self.timer.stop()
        self.is_playing = False
        self.play_button.setStyleSheet("")
        self.play_button.setText("Play")
        self.record_checkbox.setEnabled(True)
        self.frame_list.setCurrentRow(len(self.key_frames) - 1)
        if self.is_recording:
            self.stop_recording()

    def play_frames(self):
        """
        callback function for the timer to play the frames
        """
        # play the frames
        if self.current_frame_index < len(self.frames):
            key_id, Tcw = self.frames[self.current_frame_index]
            self.glwidget.set_view_matrix(np.linalg.inv(Tcw))
            self.frame_list.setCurrentRow(key_id)
            self.current_frame_index += 1
            if self.is_recording:
                self.record_frame()
        else:
            self.stop_playback()

    def toggle_recording(self, state):
        if state == 2:
            self.is_recording = True
        else:
            self.is_recording = False

    def start_recording(self):
        self.is_recording = True
        self.frames_to_record = []
        video_path = self.video_path_edit.text()
        codec = self.codec_combo.currentText()
        self.play_button.setStyleSheet("background-color: red")
        self.play_button.setText("Recording")
        self.writer = imageio.get_writer(video_path, 
                                         fps=self.update_interval,
                                         codec=codec)  # quality=10
        # disable the all the frame_item while recording
        for frame in self.key_frames:
            frame.item.hide()
        self.dock.hide()  # Hide the dock while recording

    def stop_recording(self, save_movie=True):
        self.is_recording = False
        self.record_checkbox.setChecked(False)
        # enable the all the frame_item after recording
        for frame in self.key_frames:
            frame.item.show()
        if hasattr(self, 'writer') and save_movie:
            self.writer.close()
            self.show_save_message()
        self.dock.show()  # Show the dock when recording stops

    def show_save_message(self):
        msg_box = QMessageBox()
        msg_box.setIcon(QMessageBox.Information)
        msg_box.setWindowTitle("Video Saved")
        msg_box.setText(f"Video saved to {self.video_path_edit.text()}")
        msg_box.setStandardButtons(QMessageBox.Ok)
        msg_box.exec()

    def record_frame(self):
        frame = self.glwidget.capture_frame()
        # make sure the frame size is multiple of 16
        height, width, _ = frame.shape
        if height % 16 != 0 or width % 16 != 0:
            frame = frame[:-(height % 16), :-(width % 16), :]
            frame = np.ascontiguousarray(frame)
        self.frames_to_record.append(frame)
        try:
            self.writer.append_data(frame)
        except Exception as e:
            print("Don't change the window size during recording.")
            self.stop_recording(False) # Stop recording without saving
            self.stop_playback()

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.KeyPress:
            if event.key() == QtCore.Qt.Key_Delete:
                self.del_key_frame()
                return True
        return super().eventFilter(obj, event)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Overwrite the drop event to open the cloud file.
        """
        self.progress_dialog = ProgressDialog(self)
        self.progress_dialog.show()
        files = event.mimeData().urls()
        self.progress_thread = FileLoaderThread(self, files)
        self['cloud'].load(files[0].toLocalFile(), append=False)
        self.progress_thread.progress.connect(self.file_loading_progress)
        self.progress_thread.finished.connect(self.file_loading_finished)
        self.progress_thread.start()

    def file_loading_progress(self, value):
        self.progress_dialog.set_value(value)

    def file_loading_finished(self):
        self.progress_dialog.close()

    def open_cloud_file(self, file, append=False):
        cloud_item = self['cloud']
        if cloud_item is None:
            print("Can't find clouditem.")
            return
        cloud = cloud_item.load(file, append=append)
        center = np.nanmean(cloud['xyz'].astype(np.float64), axis=0)
        self.glwidget.set_cam_position(center=center)

def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--path", help="the cloud file path")
    args = parser.parse_args()
    app = q3d.QApplication(['Film Maker'])
    viewer = CMMViewer(name='Film Maker', update_interval=30)
    cloud_item = q3d.CloudIOItem(size=0.1, point_type='SPHERE', alpha=0.5, depth_test=True)
    grid_item = q3d.GridItem(size=1000, spacing=20)

    viewer.add_items(
        {'cloud': cloud_item, 'grid': grid_item})

    if args.path:
        pcd_fn = args.path
        viewer.open_cloud_file(pcd_fn)

    viewer.show()
    app.exec()


if __name__ == '__main__':
    main()



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
import meshio
from pypcd4 import PointCloud, MetaData
from pye57 import E57
import laspy


def save_ply(cloud, save_path):
    xyz = cloud['xyz']
    i = (cloud['irgb'] & 0xFF000000) >> 24
    rgb = cloud['irgb'] & 0x00FFFFFF
    mesh = meshio.Mesh(points=xyz, cells=[], point_data={
                       "rgb": rgb, "intensity": i})
    mesh.write(save_path, file_format="ply")


def load_ply(file):
    mesh = meshio.read(file)
    xyz = mesh.points
    rgb = np.zeros([xyz.shape[0]], dtype=np.uint32)
    intensity = np.zeros([xyz.shape[0]], dtype=np.uint32)
    color_mode = 'FLAT'
    if "intensity" in mesh.point_data:
        intensity = mesh.point_data["intensity"]
        color_mode = 'I'
    if "rgb" in mesh.point_data:
        rgb = mesh.point_data["rgb"]
        color_mode = 'RGB'
    irgb = (intensity << 24) | rgb
    dtype = [('xyz', '<f4', (3,)), ('irgb', '<u4')]
    cloud = np.rec.fromarrays([xyz, irgb], dtype=dtype)
    return cloud, color_mode


def save_pcd(cloud, save_path):
    fields = ('x', 'y', 'z', 'intensity', 'rgb')
    metadata = MetaData.model_validate(
        {
            "fields": fields,
            "size": [4, 4, 4, 4, 4],
            "type": ['F', 'F', 'F', 'U', 'U'],
            "count": [1, 1, 1, 1, 1],
            "width": cloud.shape[0],
            "points": cloud.shape[0],
        })
    i = (cloud['irgb'] & 0xFF000000) >> 24
    rgb = cloud['irgb'] & 0x00FFFFFF

    dtype = [('xyz', '<f4', (3,)), ('intensity', '<u4'), ('rgb', '<u4')]
    tmp = np.rec.fromarrays([cloud['xyz'], i, rgb], dtype=dtype)

    PointCloud(metadata, tmp).save(save_path)


def load_pcd(file):
    dtype = [('xyz', '<f4', (3,)), ('irgb', '<u4')]
    pc = PointCloud.from_path(file).pc_data
    rgb = np.zeros([pc.shape[0]], dtype=np.uint32)
    intensity = np.zeros([pc.shape[0]], dtype=np.uint32)
    color_mode = 'FLAT'
    if 'intensity' in pc.dtype.names:
        intensity = pc['intensity'].astype(np.uint32)
        color_mode = 'I'
    if 'rgb' in pc.dtype.names:
        rgb = pc['rgb'].astype(np.uint32)
        color_mode = 'RGB'
    irgb = (intensity << 24) | rgb
    xyz = np.stack([pc['x'], pc['y'], pc['z']], axis=1)
    cloud = np.rec.fromarrays([xyz, irgb], dtype=dtype)
    return cloud, color_mode


def save_e57(cloud, save_path):
    e57 = E57(save_path, mode='w')
    x = cloud['xyz'][:, 0]
    y = cloud['xyz'][:, 1]
    z = cloud['xyz'][:, 2]
    i = (cloud['irgb'] & 0xFF000000) >> 24
    r = (cloud['irgb'] & 0x00FF0000) >> 16
    g = (cloud['irgb'] & 0x0000FF00) >> 8
    b = (cloud['irgb'] & 0x000000ff)
    data = {"cartesianX": x, "cartesianY": y, "cartesianZ": z,
            "intensity": i,
            "colorRed": r, "colorGreen": g, "colorBlue": b}
    e57.write_scan_raw(data)
    e57.close()


def load_e57(file_path):
    e57 = E57(file_path, mode="r")
    scans = e57.read_scan(0, ignore_missing_fields=True,
                          intensity=True, colors=True)
    x = scans["cartesianX"]
    y = scans["cartesianY"]
    z = scans["cartesianZ"]
    rgb = np.zeros([x.shape[0]], dtype=np.uint32)
    intensity = np.zeros([x.shape[0]], dtype=np.uint32)
    color_mode = 'FLAT'
    if "intensity" in scans:
        intensity = scans["intensity"].astype(np.uint32)
        color_mode = 'I'
    if all([x in scans for x in ["colorRed", "colorGreen", "colorBlue"]]):
        r = scans["colorRed"].astype(np.uint32)
        g = scans["colorGreen"].astype(np.uint32)
        b = scans["colorBlue"].astype(np.uint32)
        rgb = (r << 16) | (g << 8) | b
        color_mode = 'RGB'
    irgb = (intensity << 24) | rgb
    dtype = [('xyz', '<f4', (3,)), ('irgb', '<u4')]
    cloud = np.rec.fromarrays(
        [np.stack([x, y, z], axis=1), irgb],
        dtype=dtype)
    e57.close()
    return cloud, color_mode


def load_las(file):
    with laspy.open(file) as f:
        las = f.read()
        xyz = np.vstack((las.x, las.y, las.z)).transpose()
        dimensions = list(las.point_format.dimension_names)
        color_mode = 'FLAT'
        rgb = np.zeros([las.x.shape[0]], dtype=np.uint32)
        intensity = np.zeros([las.x.shape[0]], dtype=np.uint32)
        if 'intensity' in dimensions:
            intensity = las.intensity.astype(np.uint32)
            color_mode = 'I'
        if 'red' in dimensions and 'green' in dimensions and 'blue' in dimensions:
            red = las.red
            green = las.green
            blue = las.blue
            max_val = np.max([red, green, blue])
            if red.dtype == np.dtype('uint16') and max_val > 255:
                red = (red / 65535.0 * 255).astype(np.uint32)
                green = (green / 65535.0 * 255).astype(np.uint32)
                blue = (blue / 65535.0 * 255).astype(np.uint32)
            rgb = (red << 16) | (green << 8) | blue
            color_mode = 'RGB'
        color = (intensity << 24) | rgb
        dtype = [('xyz', '<f4', (3,)), ('irgb', '<u4')]
        cloud = np.rec.fromarrays([xyz, color], dtype=dtype)
    return cloud, color_mode

def save_las(cloud, save_path):
    header = laspy.LasHeader(point_format=3, version="1.2")
    las = laspy.LasData(header)
    las.x = cloud['xyz'][:, 0]
    las.y = cloud['xyz'][:, 1]
    las.z = cloud['xyz'][:, 2]
    las.red = (cloud['irgb'] >> 16) & 0xFF
    las.green = (cloud['irgb'] >> 8) & 0xFF
    las.blue = cloud['irgb'] & 0xFF
    las.intensity = cloud['irgb'] >> 24
    las.write(save_path)


def gsdata_type(sh_dim):
    return [('pw', '<f4', (3,)),
            ('rot', '<f4', (4,)),
            ('scale', '<f4', (3,)),
            ('alpha', '<f4'),
            ('sh', '<f4', (sh_dim))]


def matrix_to_quaternion_wxyz(matrices):
    m00, m01, m02 = matrices[:, 0, 0], matrices[:, 0, 1], matrices[:, 0, 2]
    m10, m11, m12 = matrices[:, 1, 0], matrices[:, 1, 1], matrices[:, 1, 2]
    m20, m21, m22 = matrices[:, 2, 0], matrices[:, 2, 1], matrices[:, 2, 2]
    t = 1 + m00 + m11 + m22
    s = np.ones_like(m00)
    w = np.ones_like(m00)
    x = np.ones_like(m00)
    y = np.ones_like(m00)
    z = np.ones_like(m00)

    t_positive = t > 0.0000001
    s[t_positive] = 0.5 / np.sqrt(t[t_positive])
    w[t_positive] = 0.25 / s[t_positive]
    x[t_positive] = (m21[t_positive] - m12[t_positive]) * s[t_positive]
    y[t_positive] = (m02[t_positive] - m20[t_positive]) * s[t_positive]
    z[t_positive] = (m10[t_positive] - m01[t_positive]) * s[t_positive]

    c1 = np.logical_and(m00 > m11, m00 > m22)
    cond1 = np.logical_and(np.logical_not(t_positive),
                           np.logical_and(m00 > m11, m00 > m22))

    s[cond1] = 2.0 * np.sqrt(1.0 + m00[cond1] - m11[cond1] - m22[cond1])
    w[cond1] = (m21[cond1] - m12[cond1]) / s[cond1]
    x[cond1] = 0.25 * s[cond1]
    y[cond1] = (m01[cond1] + m10[cond1]) / s[cond1]
    z[cond1] = (m02[cond1] + m20[cond1]) / s[cond1]

    c2 = np.logical_and(np.logical_not(c1), m11 > m22)
    cond2 = np.logical_and(np.logical_not(t_positive), c2)
    s[cond2] = 2.0 * np.sqrt(1.0 + m11[cond2] - m00[cond2] - m22[cond2])
    w[cond2] = (m02[cond2] - m20[cond2]) / s[cond2]
    x[cond2] = (m01[cond2] + m10[cond2]) / s[cond2]
    y[cond2] = 0.25 * s[cond2]
    z[cond2] = (m12[cond2] + m21[cond2]) / s[cond2]

    c3 = np.logical_and(np.logical_not(c1), np.logical_not(c2))
    cond3 = np.logical_and(np.logical_not(t_positive), c3)
    s[cond3] = 2.0 * np.sqrt(1.0 + m22[cond3] - m00[cond3] - m11[cond3])
    w[cond3] = (m10[cond3] - m01[cond3]) / s[cond3]
    x[cond3] = (m02[cond3] + m20[cond3]) / s[cond3]
    y[cond3] = (m12[cond3] + m21[cond3]) / s[cond3]
    z[cond3] = 0.25 * s[cond3]
    return np.array([w, x, y, z]).T


def load_gs_ply(path, T=None):
    mesh = meshio.read(path)
    vertices = mesh.points
    pws = vertices[:, :3]

    alphas = mesh.point_data['opacity']
    alphas = 1 / (1 + np.exp(-alphas))

    scales = np.vstack((mesh.point_data['scale_0'],
                        mesh.point_data['scale_1'],
                        mesh.point_data['scale_2'])).T

    rots = np.vstack((mesh.point_data['rot_0'],
                      mesh.point_data['rot_1'],
                      mesh.point_data['rot_2'],
                      mesh.point_data['rot_3'])).T
    rots /= np.linalg.norm(rots, axis=1)[:, np.newaxis]

    sh_dim = len(mesh.point_data) - 11
    shs = np.zeros([pws.shape[0], sh_dim])
    shs[:, 0] = mesh.point_data['f_dc_0']
    shs[:, 1] = mesh.point_data['f_dc_1']
    shs[:, 2] = mesh.point_data['f_dc_2']

    sh_rest_dim = sh_dim - 3
    if sh_rest_dim > 0:
        for i in range(sh_rest_dim):
            name = f"f_rest_{i}"
            shs[:, 3 + i] = mesh.point_data[name]
        shs[:, 3:] = shs[:, 3:].reshape(-1, 3, sh_rest_dim // 3).transpose([0, 2, 1]).reshape(-1, sh_rest_dim)

    pws = pws.astype(np.float32)
    rots = rots.astype(np.float32)
    scales = np.exp(scales).astype(np.float32)
    alphas = alphas.astype(np.float32)
    shs = shs.astype(np.float32)

    dtypes = gsdata_type(sh_dim)

    gs = np.rec.fromarrays(
        [pws, rots, scales, alphas, shs], dtype=dtypes)

    return gs


def rotate_gaussian(T, gs):
    # Transform to world
    pws = (T @ gs['pw'].T).T
    w = gs['rot'][:, 0]
    x = gs['rot'][:, 1]
    y = gs['rot'][:, 2]
    z = gs['rot'][:, 3]
    R = np.array([
        [1.0 - 2*(y**2 + z**2), 2*(x*y - z*w), 2*(x * z + y * w)],
        [2*(x*y + z*w), 1.0 - 2*(x**2 + z**2), 2*(y*z - x*w)],
        [2*(x*z - y*w), 2*(y*z + x*w), 1.0 - 2*(x**2 + y**2)]
    ]).transpose(2, 0, 1)
    R_new = T @ R
    rots = matrix_to_quaternion_wxyz(R_new)
    gs['pw'] = pws
    gs['rot'] = rots
    return gs


def load_gs(fn):
    if fn.endswith('.ply'):
        return load_gs_ply(fn)
    elif fn.endswith('.npy'):
        return np.load(fn)
    else:
        print("%s is not a supported file." % fn)
        exit(0)


def save_gs(fn, gs):
    np.save(fn, gs)


def get_example_gs():
    gs_data = np.array([[0.,  0.,  0.,  # xyz
                         1.,  0.,  0., 0.,  # rot
                         0.05,  0.05,  0.05,  # size
                         1.,
                         1.772484,  -1.772484,  1.772484],
                        [1.,  0.,  0.,
                         1.,  0.,  0., 0.,
                         0.2,  0.05,  0.05,
                         1.,
                         1.772484,  -1.772484, -1.772484],
                        [0.,  1.,  0.,
                         1.,  0.,  0., 0.,
                         0.05,  0.2,  0.05,
                         1.,
                         -1.772484, 1.772484, -1.772484],
                        [0.,  0.,  1.,
                         1.,  0.,  0., 0.,
                         0.05,  0.05,  0.2,
                         1.,
                         -1.772484, -1.772484,  1.772484]
                        ], dtype=np.float32)
    dtypes = gsdata_type(3)
    gs = np.frombuffer(gs_data.tobytes(), dtype=dtypes)
    return gs


if __name__ == "__main__":
    gs = load_gs("/home/liu/tmp.ply")
    print(gs.shape)



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
import numpy as np
import os
from PySide6.QtWidgets import QComboBox, QLabel
from OpenGL.GL import shaders
from q3dviewer.utils import *


def div_round_up(x, y):
    return int((x + y - 1) / y)


class GaussianItem(BaseItem):
    def __init__(self, **kwds):
        super().__init__()
        self.need_updateGS = False
        self.sh_dim = 0
        self.gs_data = np.empty([0])
        self.prev_Rz = np.array([np.inf, np.inf, np.inf])
        self.path = os.path.dirname(__file__)
        try:
            import torch
            if not torch.cuda.is_available():
                raise ImportError
            self.cuda_pw = None
            self.sort = self.torch_sort
        except ImportError:
            self.sort = self.openg_sort

    def add_setting(self, layout):
        label_render_mode = QLabel("Render Mode:")
        layout.addWidget(label_render_mode)
        combo = QComboBox()
        combo.addItem("render normal guassian")
        combo.addItem("render ball")
        combo.addItem("render inverse guassian")
        combo.currentIndexChanged.connect(self.onComboboxSelection)
        layout.addWidget(combo)

    def onComboboxSelection(self, index):
        glUseProgram(self.program)
        set_uniform(self.program, index, 'render_mod')
        glUseProgram(0)

    def initialize_gl(self):
        fragment_shader = open(
            self.path + '/../shaders/gau_frag.glsl', 'r').read()
        vertex_shader = open(
            self.path + '/../shaders/gau_vert.glsl', 'r').read()
        sort_shader = open(
            self.path + '/../shaders/sort_by_key.glsl', 'r').read()
        prep_shader = open(self.path + '/../shaders/gau_prep.glsl', 'r').read()

        self.sort_program = shaders.compileProgram(
            shaders.compileShader(sort_shader, GL_COMPUTE_SHADER))

        self.prep_program = shaders.compileProgram(
            shaders.compileShader(prep_shader, GL_COMPUTE_SHADER))

        self.program = shaders.compileProgram(
            shaders.compileShader(vertex_shader, GL_VERTEX_SHADER),
            shaders.compileShader(fragment_shader, GL_FRAGMENT_SHADER),
        )
        self.vao = glGenVertexArrays(1)

        # trade a gaussian as a square (4 2d points)
        square_vert = np.array([-1, 1, 1, 1, 1, -1, -1, -1], dtype=np.float32)
        indices = np.array([0, 1, 2, 0, 2, 3], dtype=np.uint32)

        # set the vertices for square
        vbo = glGenBuffers(1)
        glBindVertexArray(self.vao)
        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        glBufferData(GL_ARRAY_BUFFER, square_vert.nbytes,
                     square_vert, GL_STATIC_DRAW)
        pos = glGetAttribLocation(self.program, 'vert')
        glVertexAttribPointer(pos, 2, GL_FLOAT, False, 0, None)
        glEnableVertexAttribArray(pos)
        glBindBuffer(GL_ARRAY_BUFFER, 0)

        # the vert's indices for drawing square
        self.ebo = glGenBuffers(1)
        glBindVertexArray(self.vao)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.ebo)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                     indices.nbytes, indices, GL_STATIC_DRAW)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)
        glBindVertexArray(0)

        # add SSBO for gaussian data
        self.ssbo_gs = glGenBuffers(1)
        self.ssbo_gi = glGenBuffers(1)
        self.ssbo_dp = glGenBuffers(1)
        self.ssbo_pp = glGenBuffers(1)

        width = self.glwidget().current_width()
        height = self.glwidget().current_height()

        # set constant parameter for gaussian shader
        project_matrix = self.glwidget().get_projection_matrix()
        focal_x = project_matrix[0, 0] * width / 2
        focal_y = project_matrix[1, 1] * height / 2
        glUseProgram(self.prep_program)
        set_uniform(self.prep_program,
                    project_matrix, 'projection_matrix')
        set_uniform(self.prep_program, np.array([focal_x, focal_y]), 'focal')
        glUseProgram(0)

        glUseProgram(self.program)
        set_uniform(self.program, np.array([width, height]), 'win_size')
        set_uniform(self.program, 0, 'render_mod')
        glUseProgram(0)

        # opengl settings
        glDisable(GL_CULL_FACE)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def updateGS(self):
        if (self.need_updateGS):
            # compute sorting size
            self.num_sort = int(2**np.ceil(np.log2(self.gs_data.shape[0])))

            # set input gaussian data
            glBindBuffer(GL_SHADER_STORAGE_BUFFER, self.ssbo_gs)
            glBufferData(GL_SHADER_STORAGE_BUFFER, self.gs_data.nbytes,
                         self.gs_data.reshape(-1), GL_STATIC_DRAW)
            glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, self.ssbo_gs)
            glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0)

            # set depth for sorting
            glBindBuffer(GL_SHADER_STORAGE_BUFFER, self.ssbo_dp)
            glBufferData(GL_SHADER_STORAGE_BUFFER,
                         self.num_sort * 4, None, GL_STATIC_DRAW)
            glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, self.ssbo_dp)
            glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0)

            # set index for sorting (the index need be initialized)
            gi = np.arange(self.num_sort, dtype=np.uint32)
            glBindBuffer(GL_SHADER_STORAGE_BUFFER, self.ssbo_gi)
            glBufferData(GL_SHADER_STORAGE_BUFFER,
                         self.num_sort * 4, gi, GL_STATIC_DRAW)
            glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, self.ssbo_gi)
            glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0)

            # set preprocess buffer
            # the dim of preprocess data is 12 u(3),
            # covinv(3), color(3), area(2), alpha(1)
            glBindBuffer(GL_SHADER_STORAGE_BUFFER, self.ssbo_pp)
            glBufferData(GL_SHADER_STORAGE_BUFFER,
                         self.gs_data.shape[0] * 4 * 12,
                         None, GL_STATIC_DRAW)
            glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 3, self.ssbo_pp)
            glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0)

            glUseProgram(self.prep_program)
            set_uniform(self.prep_program, self.sh_dim, 'sh_dim')
            set_uniform(self.prep_program,
                        self.gs_data.shape[0], 'gs_num')
            glUseProgram(0)
            self.need_updateGS = False

    def paint(self):
        # get current view matrix
        self.view_matrix = self.glwidget().view_matrix

        # if gaussian data is update, renew vao, ssbo, etc...
        self.updateGS()

        if (self.gs_data.shape[0] == 0):
            return

        # preprocess and sort gaussian by compute shader.
        self.preprocessGS()
        self.try_sort()
        glEnable(GL_BLEND)
        # draw by vert shader
        glUseProgram(self.program)
        # bind vao and ebo
        glBindVertexArray(self.vao)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.ebo)
        # draw instances
        glDrawElementsInstanced(
            GL_TRIANGLES, 6, GL_UNSIGNED_INT, None, self.gs_data.shape[0])
        # upbind vao and ebo
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)
        glBindVertexArray(0)
        glUseProgram(0)
        glDisable(GL_BLEND)

    def try_sort(self):
        # don't sort if the depths are not change.
        Rz = self.view_matrix[2, :3]
        if (np.linalg.norm(self.prev_Rz - Rz) > 0.1):
            # import torch
            # torch.cuda.synchronize()
            # start = time.time()
            self.sort()
            self.prev_Rz = Rz
            # torch.cuda.synchronize()
            # end = time.time()
            # time_diff = end - start
            # print(time_diff)

    def openg_sort(self):
        glUseProgram(self.sort_program)
        # can we move this loop to gpu?
        # level = level*2
        for level in 2**np.arange(1, int(np.ceil(np.log2(self.num_sort))+1)):
            # stage =stage / 2
            for stage in level/2**np.arange(1, np.log2(level)+1):
                set_uniform(self.sort_program, int(level), 'level')
                set_uniform(self.sort_program, int(stage), 'stage')
                glDispatchCompute(div_round_up(self.num_sort//2, 256), 1, 1)
                glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT)
        # glFinish()
        glUseProgram(0)

    def torch_sort(self):
        import torch
        if self.cuda_pw is None:
            self.cuda_pw = torch.tensor(self.gs_data[:, :3]).cuda()
        Rz = torch.tensor(self.view_matrix[2, :3].astype(np.float32)).cuda()
        depth = Rz @ self.cuda_pw.T
        index = torch.argsort(depth).type(torch.int32).cpu().numpy()
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, self.ssbo_gi)
        glBufferData(GL_SHADER_STORAGE_BUFFER,
                     index.nbytes, index, GL_STATIC_DRAW)
        glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, self.ssbo_gi)
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0)
        return index

    def preprocessGS(self):
        glUseProgram(self.prep_program)
        set_uniform(self.prep_program, self.view_matrix, 'view_matrix')
        glDispatchCompute(div_round_up(self.gs_data.shape[0], 256), 1, 1)
        glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT)
        glUseProgram(0)

    def set_data(self, **kwds):
        if 'gs_data' in kwds:
            self.need_updateGS = False
            gs_data = kwds.pop('gs_data')
            self.gs_data = np.ascontiguousarray(gs_data, dtype=np.float32)
            self.sh_dim = self.gs_data.shape[-1] - (3 + 4 + 3 + 1)
            self.prev_Rz = np.array([np.inf, np.inf, np.inf])
            self.cuda_pw = None
            self.need_updateGS = True



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""


import numpy as np
from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
import threading
import os
from PySide6.QtWidgets import QLabel, QLineEdit, QDoubleSpinBox, \
    QComboBox, QCheckBox
from OpenGL.GL import shaders
from q3dviewer.utils import *
from q3dviewer.utils.range_slider import RangeSlider
from PySide6.QtCore import QRegularExpression
from PySide6.QtGui import QRegularExpressionValidator


# draw points with color (x, y, z, color)
class CloudItem(BaseItem):
    def __init__(self, size, alpha, 
                 color_mode='I', 
                 color='#ffffff', 
                 point_type='PIXEL', 
                 depth_test=False):
        super().__init__()
        self.STRIDE = 16  # stride of cloud array
        self.valid_buff_top = 0
        self.add_buff_loc = 0
        self.alpha = alpha
        self.size = size
        self.point_type = point_type
        self.mutex = threading.Lock()
        self.data_type = [('xyz', '<f4', (3,)), ('irgb', '<u4')]
        self.flat_rgb = int(color[1:], 16)
        self.mode_table = {'FLAT': 0,  'I': 1,  'RGB': 2}
        self.point_type_table = {'PIXEL': 0, 'SQUARE': 1, 'SPHERE': 2}
        self.color_mode = self.mode_table[color_mode]
        self.CAPACITY = 10000000  # 10MB * 3 (x,y,z, color) * 4
        self.vmin = 0
        self.vmax = 255
        self.buff = np.empty((0), self.data_type)
        self.wait_add_data = None
        self.need_update_setting = True
        self.max_cloud_size = 300000000
        # Enable depth test when full opaque
        self.depth_test = depth_test
        self.path = os.path.dirname(__file__)

    def add_setting(self, layout):
        label_ptype = QLabel("Point Type:")
        layout.addWidget(label_ptype)
        combo_ptype = QComboBox()
        combo_ptype.addItem("pixels")
        combo_ptype.addItem("flat squares")
        combo_ptype.addItem("spheres")
        combo_ptype.setCurrentIndex(self.point_type_table[self.point_type])
        combo_ptype.currentIndexChanged.connect(self._on_point_type_selection)
        layout.addWidget(combo_ptype)

        self.box_size = QDoubleSpinBox()
        self.box_size.setPrefix("Size: ")
        self.box_size.setSingleStep(1)
        self.box_size.setDecimals(0)
        self.box_size.setValue(self.size)
        self.box_size.setRange(0, 100)
        self.box_size.valueChanged.connect(self.set_size)
        self._on_point_type_selection(self.point_type_table[self.point_type])
        layout.addWidget(self.box_size)

        box_alpha = QDoubleSpinBox()
        box_alpha.setPrefix("Alpha: ")
        box_alpha.setSingleStep(0.01)
        box_alpha.setValue(self.alpha)
        box_alpha.setRange(0, 1)
        box_alpha.valueChanged.connect(self.set_alpha)
        layout.addWidget(box_alpha)

        label_color = QLabel("Color Mode:")
        layout.addWidget(label_color)
        self.combo_color = QComboBox()
        self.combo_color.addItem("flat color")
        self.combo_color.addItem("intensity")
        self.combo_color.addItem("RGB")
        self.combo_color.setCurrentIndex(self.color_mode)
        self.combo_color.currentIndexChanged.connect(self._on_color_mode)
        layout.addWidget(self.combo_color)

        label_rgb = QLabel("Color:")
        layout.addWidget(label_rgb)
        self.edit_rgb = QLineEdit()
        self.edit_rgb.setToolTip("Hex number, i.e. #FF4500")
        self.edit_rgb.setText(f"#{self.flat_rgb:06x}")
        self.edit_rgb.textChanged.connect(self._on_color)
        regex = QRegularExpression(r"^#[0-9A-Fa-f]{6}$")
        validator = QRegularExpressionValidator(regex)
        self.edit_rgb.setValidator(validator)
        layout.addWidget(self.edit_rgb)

        self.slider_v = RangeSlider()
        self.slider_v.setRange(0, 255)
        self.slider_v.rangeChanged.connect(self._on_range)
        layout.addWidget(self.slider_v)

        self.checkbox_depth_test = QCheckBox(
            "Show front points first (Depth Test)")
        self.checkbox_depth_test.setChecked(self.depth_test)
        self.checkbox_depth_test.stateChanged.connect(self.set_depthtest)
        layout.addWidget(self.checkbox_depth_test)

    def _on_range(self, lower, upper):
        self.vmin = lower
        self.vmax = upper
        self.need_update_setting = True

    def _on_color_mode(self, index):
        self.color_mode = index
        self.edit_rgb.hide()
        self.slider_v.hide()
        if (index == self.mode_table['FLAT']):  # flat color
            self.edit_rgb.show()
        elif (index == self.mode_table['I']):  # flat color
            self.slider_v.show()
        self.need_update_setting = True

    def set_color_mode(self, color_mode):
        if color_mode in {'FLAT', 'RGB', 'I'}:
            try:
                self.combo_color.setCurrentIndex(self.mode_table[color_mode])
            except RuntimeError:
                pass
            except ValueError:
                pass
        else:
            print(f"Invalid color mode: {color_mode}")

    def _on_point_type_selection(self, index):
        self.point_type = list(self.point_type_table.keys())[index]
        if self.point_type == 'PIXEL':
            self.box_size.setPrefix("Set size (pixel): ")
            self.box_size.setDecimals(0)
            self.box_size.setSingleStep(1)
            self.box_size.setValue(1)
            self.size = 1
        else:
            self.box_size.setPrefix("Set size (meter): ")
            self.box_size.setDecimals(2)
            self.box_size.setSingleStep(0.01)
            self.box_size.setValue(0.01)
            self.size = 0.01
        self.need_update_setting = True

    def set_alpha(self, alpha):
        self.alpha = alpha
        self.need_update_setting = True

    def set_flat_rgb(self, color):
        try:
            self.edit_rgb.setText(color)
        except ValueError:
            pass

    def _on_color(self, color):
        try:
            flat_rgb = int(color[1:], 16)
            self.flat_rgb = flat_rgb
            self.need_update_setting = True
        except ValueError:
            pass

    def set_size(self, size):
        self.size = size
        self.need_update_setting = True

    def set_depthtest(self, state):
        self.depth_test = state

    def clear(self):
        data = np.empty((0), self.data_type)
        self.set_data(data)

    def set_data(self, data, append=False):
        if not isinstance(data, np.ndarray):
            raise ValueError("Input data must be a numpy array.")

        if data.dtype in {np.dtype('float32'), np.dtype('float64')}:
            if data.size == 0:
                data = np.empty((0), self.data_type)
            elif data.ndim == 2 and data.shape[1] >= 3:
                xyz = data[:, :3]
                if data.shape[1] >= 4:
                    color = data[:, 3].view(np.uint32)
                else:
                    color = np.zeros(data.shape[0], dtype=np.uint32)
                data = np.rec.fromarrays(
                    [xyz, color[:data.shape[0]]], dtype=self.data_type)

        with self.mutex:
            if append:
                if self.wait_add_data is None:
                    self.wait_add_data = data
                else:
                    self.wait_add_data = np.concatenate(
                        [self.wait_add_data, data])
                self.add_buff_loc = self.valid_buff_top
            else:
                self.wait_add_data = data
                self.add_buff_loc = 0


    def update_setting(self):
        if (self.need_update_setting is False):
            return
        glUseProgram(self.program)
        set_uniform(self.program, int(self.flat_rgb), 'flat_rgb')
        set_uniform(self.program, int(self.color_mode), 'color_mode')
        set_uniform(self.program, float(self.vmax), 'vmax')
        set_uniform(self.program, float(self.vmin), 'vmin')
        set_uniform(self.program, float(self.alpha), 'alpha')
        set_uniform(self.program, float(self.size), 'point_size')
        set_uniform(self.program, int(self.point_type_table[self.point_type]), 'point_type')
        glUseProgram(0)
        self.need_update_setting = False

    def update_render_buffer(self):
        # Ensure there is data waiting to be added to the buffer
        if (self.wait_add_data is None):
            return
        # Acquire lock to update the buffer safely
        self.mutex.acquire()

        new_buff_top = self.add_buff_loc + self.wait_add_data.shape[0]
        if new_buff_top > self.buff.shape[0]:
            # if need to update buff capacity, create new cpu buff and new vbo
            buff_capacity = self.buff.shape[0]
            while (new_buff_top > buff_capacity):
                buff_capacity += self.CAPACITY
            print("[Cloud Item] Update capacity to %d" % buff_capacity)
            new_buff = np.empty((buff_capacity), self.data_type)
            new_buff[:self.add_buff_loc] = self.buff[:self.add_buff_loc]
            new_buff[self.add_buff_loc:new_buff_top] = self.wait_add_data
            self.buff = new_buff

            # if exceed the maximum cloud size, randomly select half of the points
            exceed_flag = False
            while new_buff.shape[0] > self.max_cloud_size:
                exceed_flag = True
                new_buff_half = new_buff[:new_buff_top:2]
                new_buff_top = new_buff_half.shape[0]
                new_buff = new_buff_half
            if exceed_flag:
                print("[Cloud Item] Exceed maximum cloud size %d, reduce the data size" % self.max_cloud_size)
                self.buff = self.buff[:self.max_cloud_size]
                self.buff[:new_buff_top] = new_buff


            glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
            glBufferData(GL_ARRAY_BUFFER, self.buff.nbytes,
                         self.buff, GL_DYNAMIC_DRAW)
            glBindBuffer(GL_ARRAY_BUFFER, 0)
        else:
            self.buff[self.add_buff_loc:new_buff_top] = self.wait_add_data
            glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
            glBufferSubData(GL_ARRAY_BUFFER, self.add_buff_loc * self.STRIDE,
                            self.wait_add_data.shape[0] * self.STRIDE,
                            self.wait_add_data)
        self.valid_buff_top = new_buff_top
        self.wait_add_data = None
        self.mutex.release()

    def initialize_gl(self):
        vertex_shader = open(self.path + '/../shaders/cloud_vert.glsl', 'r').read()
        fragment_shader = open(self.path + '/../shaders/cloud_frag.glsl', 'r').read()
        self.program = shaders.compileProgram(
            shaders.compileShader(vertex_shader, GL_VERTEX_SHADER),
            shaders.compileShader(fragment_shader, GL_FRAGMENT_SHADER),
        )
        self.max_cloud_size = glGetIntegerv(
            GL_MAX_SHADER_STORAGE_BLOCK_SIZE) // self.STRIDE
        # Bind attribute locations
        self.vbo = glGenBuffers(1)

    def paint(self):
        self.update_render_buffer()
        self.update_setting()
        glEnable(GL_BLEND)
        glEnable(GL_PROGRAM_POINT_SIZE)
        glEnable(GL_POINT_SPRITE)
        if self.depth_test:
            glEnable(GL_DEPTH_TEST)
        else:
            glDisable(GL_DEPTH_TEST)

        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glUseProgram(self.program)
        glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE,
                              self.STRIDE, ctypes.c_void_p(0))
        glVertexAttribPointer(
            1, 1, GL_FLOAT, GL_UNSIGNED_INT, self.STRIDE, ctypes.c_void_p(12))
        glEnableVertexAttribArray(0)
        glEnableVertexAttribArray(1)

        view_matrix = self.glwidget().view_matrix
        set_uniform(self.program, view_matrix, 'view_matrix')
        project_matrix = self.glwidget().projection_matrix
        set_uniform(self.program, project_matrix, 'projection_matrix')
        width = self.glwidget().current_width()
        focal = project_matrix[0, 0] * width / 2
        set_uniform(self.program, float(focal), 'focal')

        glDrawArrays(GL_POINTS, 0, self.valid_buff_top)

        # unbind VBO
        glDisableVertexAttribArray(0)
        glDisableVertexAttribArray(1)
        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glUseProgram(0)
        glDisable(GL_POINT_SPRITE)
        glDisable(GL_PROGRAM_POINT_SIZE)
        glDisable(GL_BLEND)
        if self.depth_test:
            glDisable(GL_DEPTH_TEST)  # Disable depth testing if it was enabled



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
from pypcd4 import PointCloud
from q3dviewer.utils.maths import make_transform


def convert_pointcloud2_msg(msg):
    pc = PointCloud.from_msg(msg).pc_data
    data_type = [('xyz', '<f4', (3,)), ('irgb', '<u4')]
    rgb = np.zeros([pc.shape[0]], dtype=np.uint32)
    intensity = np.zeros([pc.shape[0]], dtype=np.uint32)
    fields = ['xyz']
    if 'intensity' in pc.dtype.names:
        intensity = pc['intensity'].astype(np.uint32)
        fields.append('intensity')
    if 'rgb' in pc.dtype.names:
        rgb = pc['rgb'].view(np.uint32)
        fields.append('rgb')
    irgb = (intensity << 24) | rgb
    xyz = np.stack([pc['x'], pc['y'], pc['z']], axis=1)
    cloud = np.rec.fromarrays([xyz, irgb], dtype=data_type)
    stamp = msg.header.stamp.to_sec()
    return cloud, fields, stamp


def convert_odometry_msg(msg):
    pose = np.array(
        [msg.pose.pose.position.x,
         msg.pose.pose.position.y,
         msg.pose.pose.position.z])
    rotation = np.array([
        msg.pose.pose.orientation.x,
        msg.pose.pose.orientation.y,
        msg.pose.pose.orientation.z,
        msg.pose.pose.orientation.w])
    transform = make_transform(pose, rotation)
    stamp = msg.header.stamp.to_sec()
    return transform, stamp


def convert_image_msg(msg):
    image = np.frombuffer(msg.data, dtype=np.uint8).reshape(
        msg.height, msg.width, -1)
    if (msg.encoding == 'bgr8'):
        image = image[:, :, ::-1]  # convert bgr to rgb
    stamp = msg.header.stamp.to_sec()
    return image, stamp



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
from PySide6.QtWidgets import QLabel, QDoubleSpinBox, QLineEdit
from PySide6.QtCore import QRegularExpression
from PySide6.QtGui import QRegularExpressionValidator
import numpy as np
from q3dviewer.utils.maths import hex_to_rgba


class GridItem(BaseItem):
    def __init__(self, size=100, spacing=20, color='#ffffff40', offset=np.array([0., 0., 0.])):
        super().__init__()
        self.size = size
        self.spacing = spacing
        self.offset = offset
        self.need_update_grid = True
        self.set_color(color)
        self.vertices = self.generate_grid_vertices()

    def set_color(self, color):
        try:
            self.rgba = hex_to_rgba(color)
        except ValueError:
            pass

    def generate_grid_vertices(self):
        vertices = []
        x, y, z = self.offset
        half_size = self.size / 2  # Keep as float
        for i in np.arange(-half_size, half_size + self.spacing, self.spacing):
            vertices.extend([i + x, -half_size + y, z, i + x, half_size + y, z])  # Grid lines parallel to Y axis
            vertices.extend([-half_size + x, i + y, z, half_size + x, i + y, z])  # Grid lines parallel to X axis
        vertices = np.array(vertices, dtype=np.float32)
        return vertices

    def initialize_gl(self):
        self.vao = glGenVertexArrays(1)
        vbo = glGenBuffers(1)

        glBindVertexArray(self.vao)

        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        glBufferData(GL_ARRAY_BUFFER, self.vertices.nbytes, self.vertices, GL_STATIC_DRAW)

        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, ctypes.c_void_p(0))
        glEnableVertexAttribArray(0)

        glBindVertexArray(0)

    def add_setting(self, layout):
        spinbox_size = QDoubleSpinBox()
        spinbox_size.setPrefix("Size: ")
        spinbox_size.setSingleStep(1.0)
        spinbox_size.setValue(self.size)
        spinbox_size.setRange(0, 100000)
        spinbox_size.valueChanged.connect(self.set_size)
        layout.addWidget(spinbox_size)

        spinbox_spacing = QDoubleSpinBox()
        spinbox_spacing.setPrefix("Spacing: ")
        spinbox_spacing.setSingleStep(0.1)
        spinbox_spacing.setValue(self.spacing)
        spinbox_spacing.setRange(0, 1000)
        spinbox_spacing.valueChanged.connect(self._on_spacing)
        layout.addWidget(spinbox_spacing)

        spinbox_offset_x = QDoubleSpinBox()
        spinbox_offset_x.setPrefix("Offset X: ")
        spinbox_offset_x.setSingleStep(0.1)
        spinbox_offset_x.setValue(self.offset[0])
        spinbox_offset_x.valueChanged.connect(self._on_offset_x)
        layout.addWidget(spinbox_offset_x)
        
        spinbox_offset_y = QDoubleSpinBox()
        spinbox_offset_y.setPrefix("Offset Y: ")
        spinbox_offset_y.setSingleStep(0.1)
        spinbox_offset_y.setValue(self.offset[1])
        spinbox_offset_y.valueChanged.connect(self._on_offset_y)
        layout.addWidget(spinbox_offset_y)
        
        spinbox_offset_z = QDoubleSpinBox()
        spinbox_offset_z.setPrefix("Offset Z: ")
        spinbox_offset_z.setSingleStep(0.1)
        spinbox_offset_z.setValue(self.offset[2])
        spinbox_offset_z.valueChanged.connect(self._on_offset_z)
        layout.addWidget(spinbox_offset_z)

    def set_size(self, size):
        self.size = size
        self.need_update_grid = True

    def _on_spacing(self, spacing):
        if spacing > 0:
            self.spacing = spacing
            self.need_update_grid = True

    def _on_offset_x(self, value):
        self.offset[0] = value
        print(self.offset)
        self.need_update_grid = True

    def _on_offset_y(self, value):
        self.offset[1] = value
        self.need_update_grid = True

    def _on_offset_z(self, value):
        self.offset[2] = value
        self.need_update_grid = True

    def set_offset(self, offset):
        if isinstance(offset, np.ndarray) and offset.shape == (3,):
            self.offset = offset
            self.need_update_grid = True
        else:
            raise ValueError("Offset must be a numpy array with shape (3,)")

    def paint(self):
        if self.need_update_grid:
            self.vertices = self.generate_grid_vertices()
            self.initialize_gl()
            self.need_update_grid = False
        
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glLineWidth(1)
        glColor4f(*self.rgba)
        glBindVertexArray(self.vao)
        glDrawArrays(GL_LINES, 0, len(self.vertices) // 3)
        glBindVertexArray(0)
        glLineWidth(1)
        glDisable(GL_BLEND)





#!/usr/bin/env python3
"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.glwidget import *
import signal
from PySide6.QtWidgets import QMainWindow, QApplication, QHBoxLayout


def handler(signal, frame):
    QApplication.quit()
    print("Force close by Ctrl+C")


class Viewer(QMainWindow):
    def __init__(self, name='Viewer', win_size=[1920, 1080], 
                 gl_widget_class=GLWidget, update_interval=20):
        signal.signal(signal.SIGINT, handler)
        super(Viewer, self).__init__()
        self.setGeometry(0, 0, win_size[0], win_size[1])
        self.gl_widget_class = gl_widget_class
        self.init_ui()
        self.update_interval = update_interval
        self.add_update_timer()
        self.setWindowTitle(name)
        self.installEventFilter(self)

    def init_ui(self):
        center_widget = QWidget()
        self.setCentralWidget(center_widget)
        main_layout = QHBoxLayout()
        self.add_control_panel(main_layout)
        center_widget.setLayout(main_layout)
        self.glwidget = self.gl_widget_class()
        main_layout.addWidget(self.glwidget, 1)
        self.default_gl_setting(self.glwidget)

    def add_control_panel(self, main_layout):
        """
        Override this function to add your own control panel to 
        the left side of the main window.
        Don't forget add your own layout to the main_layout.
        """
        pass

    def default_gl_setting(self, glwidget):
        """
        Override this function to set the default opengl setting of the viewer.
        """
        pass

    def add_update_timer(self):
        timer = QtCore.QTimer(self)
        timer.setInterval(self.update_interval)  # period, in milliseconds
        timer.timeout.connect(self.update)
        timer.start()

    def add_items(self, named_items: dict):
        for name, item in named_items.items():
            self.glwidget.add_item_with_name(name, item)

    def __getitem__(self, name: str):
        if name in self.glwidget.named_items:
            return self.glwidget.named_items[name]
        else:
            return None

    def update(self):
        # force update by timer
        self.glwidget.update()

    def closeEvent(self, event):
        event.accept()
        QApplication.quit()

    def show(self):
        self.glwidget.setting_window.add_setting(
            "main win", self.glwidget)
        super().show()



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
import numpy as np
from OpenGL.GL import shaders
from PIL import Image as PIL_Image
from PySide6.QtWidgets import QLabel, QSpinBox, QCheckBox


# Vertex and Fragment shader source code
vertex_shader_source = """
#version 330 core
layout(location = 0) in vec3 position;
layout(location = 1) in vec2 texCoord;

out vec2 TexCoord;

uniform mat4 view_matrix;
uniform mat4 project_matrix;

void main()
{
    gl_Position = vec4(position, 1.0);
    TexCoord = texCoord;
}
"""

fragment_shader_source = """
#version 330 core
in vec2 TexCoord;
out vec4 color;
uniform sampler2D ourTexture;
void main()
{
    color = texture(ourTexture, TexCoord);
}
"""


class ImageItem(BaseItem):
    def __init__(self, pos=np.array([0, 0]), size=np.array([1280/2, 720/2])):
        BaseItem.__init__(self)
        self.pos = pos  # bottom-left
        self.size = size
        self.image = np.zeros((self.size[0], self.size[1], 4), dtype=np.uint8)
        self.alpha = 255

    def initialize_gl(self):
        # Rectangle vertices and texture coordinates
        width = self.glwidget().current_width()
        height = self.glwidget().current_height()
        x0, y0 = self.pos
        x1, y1 = self.pos + self.size
        x0 = x0 / width * 2 - 1
        y0 = y0 / height * 2 - 1
        x1 = x1 / width * 2 - 1
        y1 = y1 / height * 2 - 1

        self.vertices = np.array([
            # positions          # texture coords
            [x0, y0,  0.0,  0.0, 0.0],  # bottom-left
            [x1, y0,  0.0,  1.0, 0.0],  # bottom-right
            [x1,  y1,  0.0,  1.0, 1.0],  # top-right
            [x0,  y1,  0.0,  0.0, 1.0],   # top-left
        ], dtype=np.float32)

        indices = np.array([
            0, 1, 2,  # first triangle
            2, 3, 0   # second triangle
        ], dtype=np.uint32)

        self.vao = glGenVertexArrays(1)
        vbo = glGenBuffers(1)
        ebo = glGenBuffers(1)

        glBindVertexArray(self.vao)

        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        glBufferData(GL_ARRAY_BUFFER, self.vertices.itemsize *
                     5 * 4, self.vertices, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                     indices.nbytes, indices, GL_STATIC_DRAW)

        # Vertex positions
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE,
                              20, ctypes.c_void_p(0))
        glEnableVertexAttribArray(0)

        # Texture coordinates
        glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE,
                              20, ctypes.c_void_p(12))
        glEnableVertexAttribArray(1)

        # Compile shaders and create shader program
        self.program = shaders.compileProgram(
            shaders.compileShader(vertex_shader_source, GL_VERTEX_SHADER),
            shaders.compileShader(fragment_shader_source, GL_FRAGMENT_SHADER),
        )

        self.texture = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self.texture)
        glBindVertexArray(0)

    def set_data(self, data):
        if isinstance(data, np.ndarray):
            pass
        elif isinstance(data, PIL_Image.Image):
            data = np.array(data)
        else:
            print("not support image type")
            raise NotImplementedError

        if data.ndim == 2:  # Grayscale image
            data = np.stack((data,) * 3 + (np.ones_like(data) * 255,), axis=-1)
        elif data.shape[-1] == 3:  # RGB image
            alpha_channel = np.ones(
                (data.shape[0], data.shape[1], 1),
                dtype=data.dtype) * self.alpha
            data = np.concatenate((data, alpha_channel), axis=-1)
        self.image = data

    def paint(self):
        if self.image is not None:
            img_data = self.image
            img_data = np.flipud(img_data)  # Flip the image vertically
            img_data = img_data.tobytes()
            glBindTexture(GL_TEXTURE_2D, self.texture)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.image.shape[1],
                         self.image.shape[0], 0,
                         GL_RGBA, GL_UNSIGNED_BYTE, img_data)
            glGenerateMipmap(GL_TEXTURE_2D)
            glBindTexture(GL_TEXTURE_2D, 0)
            self.image = None

        glEnable(GL_DEPTH_TEST)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glUseProgram(self.program)
        glBindVertexArray(self.vao)
        glBindTexture(GL_TEXTURE_2D, self.texture)
        glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, None)
        glBindTexture(GL_TEXTURE_2D, 0)
        glBindVertexArray(0)
        glUseProgram(0)

        glDisable(GL_DEPTH_TEST)
        glDisable(GL_BLEND)

    def add_setting(self, layout):
        spinbox_alpha = QSpinBox()
        spinbox_alpha.setPrefix("Alpha: ")
        spinbox_alpha.setSingleStep(1)
        spinbox_alpha.setRange(0, 255)
        spinbox_alpha.setValue(self.alpha)
        spinbox_alpha.valueChanged.connect(self.set_alpha)
        layout.addWidget(spinbox_alpha)
        
        checkbox_show = QCheckBox("Show Image")
        checkbox_show.setChecked(True)
        checkbox_show.stateChanged.connect(self.set_visible)
        layout.addWidget(checkbox_show)

    def set_alpha(self, alpha):
        self.alpha = alpha



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from PySide6 import QtCore
import numpy as np

class BaseItem(QtCore.QObject):
    _next_id = 0
    
    def __init__(self):
        super().__init__()
        self._id = BaseItem._next_id
        BaseItem._next_id += 1
        self._glwidget = None
        self._visible = True
        self._initialized = False
        self._disable_setting = False
        
    def set_glwidget(self, v):
        self._glwidget = v
        
    def glwidget(self):
        return self._glwidget
    
    def hide(self):
        self._visible = False
        
    def show(self):
        self._visible = True
    
    def set_visible(self, vis):
        self._visible = vis
        
    def visible(self):
        return self._visible
    
    def initialize(self):
        if not self._initialized:
            self.initialize_gl()
            self._initialized = True

    def is_initialized(self):
        return self._initialized

    def add_setting(self, layout):
        """
        Add setting widgets to the layout.
        This method should be overridden by subclasses to add any necessary setting widgets to the layout.
        """
        pass
    
    def initialize_gl(self):
        """
        Initialize OpenGL resources for the item.
        This method should be overridden by subclasses to set up any necessary OpenGL resources.
        """
        pass

    def paint(self):
        """
        Render the item using OpenGL.
        This method should be overridden by subclasses to perform the actual rendering.
        """
        pass

    def disable_setting(self):
        self._disable_setting = True






#!/usr/bin/env python3

"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from sensor_msgs.msg import PointCloud2
import rospy
import numpy as np
import argparse
import q3dviewer as q3d
from PySide6.QtWidgets import QLabel, QLineEdit, QDoubleSpinBox, \
    QSpinBox, QWidget, QVBoxLayout, QHBoxLayout, QPushButton
from PySide6 import QtCore
from q3dviewer.utils.convert_ros_msg import convert_pointcloud2_msg

try:
    import open3d as o3d
except ImportError:
    print("\033[91mWarning: open3d is not installed. Please install it to use this tool.\033[0m")
    print("\033[93mYou can install it using: pip install open3d\033[0m")
    exit(1)


viewer = None
cloud0_accum = None
clouds0 = []
cloud1_accum = None
clouds1 = []


class CustomDoubleSpinBox(QDoubleSpinBox):
    def __init__(self, decimals=4, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.decimals = decimals
        self.setDecimals(self.decimals)

    def textFromValue(self, value):
        return f"{value:.{self.decimals}f}"

    def valueFromText(self, text):
        return float(text)


class LiDARCalibViewer(q3d.Viewer):
    def __init__(self, **kwargs):
        self.t01 = np.array([0, 0, 0])
        self.R01 = np.eye(3)
        self.cloud_num = 10
        self.radius = 0.2
        super().__init__(**kwargs)

    def default_gl_setting(self, glwidget):
        # Set camera position and background color
        glwidget.set_bg_color('#ffffff')
        glwidget.set_cam_position(distance=5)
    
    def add_control_panel(self, main_layout):
        # Create a vertical layout for the settings
        setting_layout = QVBoxLayout()
        setting_layout.setAlignment(QtCore.Qt.AlignTop)
        # Add XYZ spin boxes
        label_xyz = QLabel("Set XYZ:")
        setting_layout.addWidget(label_xyz)
        self.box_x = CustomDoubleSpinBox()
        self.box_x.setSingleStep(0.01)
        setting_layout.addWidget(self.box_x)
        self.box_y = CustomDoubleSpinBox()
        self.box_y.setSingleStep(0.01)
        setting_layout.addWidget(self.box_y)
        self.box_z = CustomDoubleSpinBox()
        self.box_z.setSingleStep(0.01)
        setting_layout.addWidget(self.box_z)
        self.box_x.setRange(-100.0, 100.0)
        self.box_y.setRange(-100.0, 100.0)
        self.box_z.setRange(-100.0, 100.0)

        # Add RPY spin boxes
        label_rpy = QLabel("Set Roll-Pitch-Yaw:")
        setting_layout.addWidget(label_rpy)
        self.box_roll = CustomDoubleSpinBox()
        self.box_roll.setSingleStep(0.01)
        setting_layout.addWidget(self.box_roll)
        self.box_pitch = CustomDoubleSpinBox()
        self.box_pitch.setSingleStep(0.01)
        setting_layout.addWidget(self.box_pitch)
        self.box_yaw = CustomDoubleSpinBox()
        self.box_yaw.setSingleStep(0.01)
        setting_layout.addWidget(self.box_yaw)
        self.box_roll.setRange(-np.pi, np.pi)
        self.box_pitch.setRange(-np.pi, np.pi)
        self.box_yaw.setRange(-np.pi, np.pi)

        label_cloud_num = QLabel("Set cloud frame number:")
        setting_layout.addWidget(label_cloud_num)
        self.box_cloud_num = QSpinBox()
        self.box_cloud_num.setValue(self.cloud_num)
        self.box_cloud_num.setRange(1, 100)
        self.box_cloud_num.valueChanged.connect(self.update_cloud_num)
        setting_layout.addWidget(self.box_cloud_num)

        label_trans_quat = QLabel("The lidar0-lidar1 trans and quat:")
        setting_layout.addWidget(label_trans_quat)
        self.line_trans = QLineEdit()
        self.line_trans.setReadOnly(True)
        setting_layout.addWidget(self.line_trans)
        self.line_quat = QLineEdit()
        self.line_quat.setReadOnly(True)
        setting_layout.addWidget(self.line_quat)

        label_matching_setting = QLabel("Set matching radius:")
        setting_layout.addWidget(label_matching_setting)
        self.box_radius = CustomDoubleSpinBox(decimals=2)
        self.box_radius.setSingleStep(0.1)
        self.box_radius.setRange(0.1, 3.0)
        self.box_radius.setValue(self.radius)
        setting_layout.addWidget(self.box_radius)

        self.icp_button = QPushButton("Auto Scan Matching")
        self.icp_button.clicked.connect(self.perform_matching)
        setting_layout.addWidget(self.icp_button)

        self.line_trans.setText(
            f"[{self.t01[0]:.6f}, {self.t01[1]:.6f}, {self.t01[2]:.6f}]")
        quat = q3d.matrix_to_quaternion(self.R01)
        self.line_quat.setText(
            f"[{quat[0]:.6f}, {quat[1]:.6f}, {quat[2]:.6f}, {quat[3]:.6f}]")

        # Connect spin boxes to methods
        self.box_x.valueChanged.connect(self.update_xyz)
        self.box_y.valueChanged.connect(self.update_xyz)
        self.box_z.valueChanged.connect(self.update_xyz)
        self.box_roll.valueChanged.connect(self.update_rpy)
        self.box_pitch.valueChanged.connect(self.update_rpy)
        self.box_yaw.valueChanged.connect(self.update_rpy)

        main_layout.addLayout(setting_layout)
        

    def update_radius(self):
        self.radius = self.box_radius.value()

    def update_cloud_num(self):
        self.cloud_num = self.box_cloud_num.value()

    def update_xyz(self):
        x = self.box_x.value()
        y = self.box_y.value()
        z = self.box_z.value()
        self.t01 = np.array([x, y, z])
        self.line_trans.setText(f"[{x:.6f}, {y:.6f}, {z:.6f}]")

    def update_rpy(self):
        roll = self.box_roll.value()
        pitch = self.box_pitch.value()
        yaw = self.box_yaw.value()
        self.R01 = q3d.euler_to_matrix(np.array([roll, pitch, yaw]))
        quat = q3d.matrix_to_quaternion(self.R01)
        self.line_quat.setText(
            f"[{quat[0]:.6f}, {quat[1]:.6f}, {quat[2]:.6f}, {quat[3]:.6f}]")

    def perform_matching(self):
        global cloud0_accum, cloud1_accum

        if o3d is None:
            print("Error: open3d is not available. Cannot perform matching.")
            return

        if cloud0_accum is not None and cloud1_accum is not None:
            # Convert to Open3D point clouds
            cloud0_o3d = o3d.geometry.PointCloud()
            cloud0_o3d.points = o3d.utility.Vector3dVector(cloud0_accum['xyz'])
            cloud1_o3d = o3d.geometry.PointCloud()
            cloud1_o3d.points = o3d.utility.Vector3dVector(cloud1_accum['xyz'])
            voxel_size = 0.1
            cloud0_o3d = cloud0_o3d.voxel_down_sample(voxel_size)
            cloud1_o3d = cloud1_o3d.voxel_down_sample(voxel_size)
            cloud0_o3d.estimate_normals(
                search_param=o3d.geometry.KDTreeSearchParamHybrid
                (radius=self.radius, max_nn=30))
            trans_init = np.eye(4)
            trans_init[:3, :3] = self.R01
            trans_init[:3, 3] = self.t01

            criteria = o3d.pipelines.registration.ICPConvergenceCriteria(
                relative_fitness=1e-4,
                max_iteration=50)

            # Auto Scan Matching
            reg_p2p = o3d.pipelines.registration.registration_icp(
                cloud1_o3d, cloud0_o3d, self.radius, trans_init,
                o3d.pipelines.registration.
                TransformationEstimationPointToPlane(),
                criteria)

            transformation_icp = reg_p2p.transformation
            # print("ICP Transformation:", transformation_icp)

            # Update R01 and t01 with ICP result
            R01 = transformation_icp[:3, :3]
            t01 = transformation_icp[:3, 3]

            # Update the UI with new values
            quat = q3d.matrix_to_quaternion(R01)
            rpy = q3d.matrix_to_euler(R01)
            self.box_roll.setValue(rpy[0])
            self.box_pitch.setValue(rpy[1])
            self.box_yaw.setValue(rpy[2])
            self.box_x.setValue(t01[0])
            self.box_y.setValue(t01[1])
            self.box_z.setValue(t01[2])
            self.line_trans.setText(
                f"[{t01[0]:.6f}, {t01[1]:.6f}, {t01[2]:.6f}]")
            self.line_quat.setText(
                f"[{quat[0]: .6f}, {quat[1]: .6f}, {quat[2]: .6f}, {quat[3]: .6f}]")
            self.t01 = t01
            self.R01 = R01
            print("Matching results")
            print(f"translation: [{t01[0]:.6f}, {t01[1]:.6f}, {t01[2]:.6f}]")
            print(
                f"Roll-Pitch-Yaw: [{rpy[0]:.6f}, {rpy[1]:.6f}, {rpy[2]:.6f}]")
            print(
                f"Quaternion: [{quat[0]: .6f}, {quat[1]: .6f}, {quat[2]: .6f}, {quat[3]: .6f}]")


def scan0_cb(data):
    global viewer, clouds0, cloud0_accum
    cloud, _, _  = convert_pointcloud2_msg(data)
    while len(clouds0) > viewer.cloud_num:
        clouds0.pop(0)
    clouds0.append(cloud)
    cloud0_accum = np.concatenate(clouds0)
    viewer['scan0'].set_data(data=cloud0_accum)


def scan1_cb(data):
    global viewer, clouds1, cloud1_accum
    cloud, _, _  = convert_pointcloud2_msg(data)
    while len(clouds1) > viewer.cloud_num:
        clouds1.pop(0)
    clouds1.append(cloud)
    cloud1_accum = np.concatenate(clouds1)
    cloud0_accum_new = cloud1_accum.copy()
    cloud0_accum_new['xyz'] = (
        viewer.R01 @ cloud1_accum['xyz'].T + viewer.t01[:, np.newaxis]).T

    viewer['scan1'].set_data(data=cloud0_accum_new)


def main():
    global viewer
    # Set up argument parser
    parser = argparse.ArgumentParser(
        description="Configure topic names for LiDAR, image, and camera info.")
    parser.add_argument('--lidar0', type=str, default='/lidar0',
                        help="Topic name for LiDAR data")
    parser.add_argument('--lidar1', type=str, default='/lidar1',
                        help="Topic name for camera image data")
    args = parser.parse_args()

    app = q3d.QApplication(["LiDAR Calib"])
    viewer = LiDARCalibViewer(name='LiDAR Calib')
    grid_item = q3d.GridItem(size=10, spacing=1, color='#00000040')
    scan0_item = q3d.CloudItem(
        size=2, alpha=1, color_mode='FLAT', color='#ff0000')
    scan1_item = q3d.CloudItem(
        size=2, alpha=1, color_mode='FLAT', color='#00ff00')
    viewer.add_items(
        {'scan0': scan0_item, 'scan1': scan1_item, 'grid': grid_item})

    rospy.init_node('lidar_calib', anonymous=True)

    # Use topic names from arguments
    rospy.Subscriber(args.lidar0, PointCloud2, scan0_cb, queue_size=1)
    rospy.Subscriber(args.lidar1, PointCloud2, scan1_cb, queue_size=1)

    viewer.show()
    app.exec()


if __name__ == '__main__':
    main()



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.custom_items.cloud_item import CloudItem
from pathlib import Path
import os
from PySide6.QtWidgets import QPushButton, QLabel, QLineEdit, QMessageBox
from q3dviewer.utils.cloud_io import save_pcd, save_ply, save_e57, save_las, load_pcd, load_ply, load_e57, load_las

class CloudIOItem(CloudItem):
    """
    add save/load function to CloudItem
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.save_path = str(Path(os.path.expanduser("~"), "data.pcd"))

    def add_setting(self, layout):
        super().add_setting(layout)

        label4 = QLabel("Save Path:")
        layout.addWidget(label4)
        box4 = QLineEdit()
        box4.setText(self.save_path)
        box4.textChanged.connect(self.set_path)
        layout.addWidget(box4)
        save_button = QPushButton("Save Cloud")
        save_button.clicked.connect(self.save)
        layout.addWidget(save_button)
        self.save_msg = QMessageBox()
        self.save_msg.setIcon(QMessageBox.Information)
        self.save_msg.setWindowTitle("save")
        self.save_msg.setStandardButtons(QMessageBox.Ok)

    def save(self):
        cloud = self.buff[:self.valid_buff_top]
        func = None
        if self.save_path.endswith(".pcd"):
            func = save_pcd
        elif self.save_path.endswith(".ply"):
            func = save_ply
        elif self.save_path.endswith(".e57"):
            func = save_e57
        elif self.save_path.endswith(".las"):
            func = save_las
        elif self.save_path.endswith(".tif") or self.save_path.endswith(".tiff"):
            print("Do not support save as tif type!")
        else:
            print("Unsupported cloud file type!")
        try:
            func(cloud, self.save_path)
            self.save_msg.setText("Save cloud to  %s" % self.save_path)
        except Exception as e:
            print(e)
            self.save_msg.setText("Cannot save to %s" % self.save_path)
            self.save_msg.exec()
        self.save_msg.exec()

    def load(self, file, append=False):
        # print("Try to load %s ..." % file)
        if file.endswith(".pcd"):
            cloud, color_mode = load_pcd(file)
        elif file.endswith(".ply"):
            cloud, color_mode = load_ply(file)
        elif file.endswith(".e57"):
            cloud, color_mode = load_e57(file)
        elif file.endswith(".las"):
            cloud, color_mode = load_las(file)
        else:
            print("Not supported file type.")
            return
        self.set_data(data=cloud, append=append)
        self.set_color_mode(color_mode)
        return cloud

    def set_path(self, path):
        self.save_path = path



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from PySide6 import QtCore
from PySide6.QtWidgets import QWidget, QComboBox, QVBoxLayout, QHBoxLayout, QLabel, QLineEdit, QCheckBox, QGroupBox
from PySide6.QtGui import QKeyEvent, QVector3D, QRegularExpressionValidator
from PySide6.QtCore import QRegularExpression
from OpenGL.GL import *
import numpy as np
from q3dviewer.base_glwidget import BaseGLWidget

class SettingWindow(QWidget):
    def __init__(self):
        super().__init__()
        self.combo_items = QComboBox()
        self.combo_items.currentIndexChanged.connect(self.on_combo_selection)
        main_layout = QVBoxLayout()
        main_layout.setAlignment(QtCore.Qt.AlignTop)
        main_layout.addWidget(self.combo_items)
        self.layout = QVBoxLayout()
        main_layout.addLayout(self.layout)
        self.setLayout(main_layout)
        self.setWindowTitle("Setting Window")
        self.setGeometry(200, 200, 300, 200)
        self.items = {}

    def add_setting(self, name, item):
        self.items.update({name: item})
        self.combo_items.addItem("%s(%s)" % (name, item.__class__.__name__))

    def clear_setting(self):
        while self.layout.count():
            child = self.layout.takeAt(0)
            if child.widget():
                child.widget().deleteLater()

    def on_combo_selection(self, index):
        # remove all setting of previous widget
        self.clear_setting()
        key = list(self.items.keys())
        item = self.items[key[index]]
        group_box = QGroupBox()
        group_layout = QVBoxLayout()
        item.add_setting(group_layout)
        group_box.setLayout(group_layout)
        self.layout.addWidget(group_box)


class GLWidget(BaseGLWidget):
    def __init__(self):
        self.followed_name = 'none'
        self.named_items = {}
        self.color_str = '#000000'
        self.followable_item_name = None
        self.setting_window = SettingWindow()
        self.enable_show_center = True
        super(GLWidget, self).__init__()

    def keyPressEvent(self, ev: QKeyEvent):
        if ev.key() == QtCore.Qt.Key_M:  # setting meun
            print("Open setting windows")
            self.open_setting_window()
        super().keyPressEvent(ev)

    def on_followable_selection(self, index):
        self.followed_name = self.followable_item_name[index]

    def update(self):
        if self.followed_name != 'none':
            new_center = self.named_items[self.followed_name].T[:3, 3]
            self.set_center(new_center)
        super().update()

    def add_setting(self, layout):
        label_color = QLabel("Set background color:")
        layout.addWidget(label_color)
        color_edit = QLineEdit()
        color_edit.setToolTip("'using hex color, i.e. #FF4500")
        color_edit.setText(self.color_str)
        color_edit.textChanged.connect(self.set_bg_color)
        regex = QRegularExpression(r"^#[0-9A-Fa-f]{6}$")
        validator = QRegularExpressionValidator(regex)
        color_edit.setValidator(validator)
        layout.addWidget(color_edit)
        
        label_focus = QLabel("Set Focus:")
        combo_focus = QComboBox()
    
        if self.followable_item_name is None:
            self.initial_followable()

        for name in self.followable_item_name:
            combo_focus.addItem(name)
        combo_focus.currentIndexChanged.connect(self.on_followable_selection)
        layout.addWidget(label_focus)
        layout.addWidget(combo_focus)

        checkbox_show_center = QCheckBox("Show Center Point")
        checkbox_show_center.setChecked(self.enable_show_center)
        checkbox_show_center.stateChanged.connect(self.change_show_center)
        layout.addWidget(checkbox_show_center)

    def initial_followable(self):
        self.followable_item_name = ['none']
        for name, item in self.named_items.items():
            if item.__class__.__name__ == 'AxisItem' and not item._disable_setting:
                self.followable_item_name.append(name)

    def set_bg_color(self, color_str):
        try:
            color_flat = int(color_str[1:], 16)
            red = (color_flat >> 16) & 0xFF
            green = (color_flat >> 8) & 0xFF
            blue = color_flat & 0xFF
            self.color_str = color_str
            self.set_color([red, green, blue, 0])
        except ValueError:
            return

    def add_item_with_name(self, name, item):
        self.named_items.update({name: item})
        if not item._disable_setting:
            self.setting_window.add_setting(name, item)
        super().add_item(item)

    def open_setting_window(self):
        if self.setting_window.isVisible():
            self.setting_window.raise_()

        else:
            self.setting_window.show()

    def change_show_center(self, state):
        self.enable_show_center = state



#!/usr/bin/env python3

"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
import q3dviewer as q3d
from PySide6.QtWidgets import QVBoxLayout, QProgressBar, QDialog, QLabel
from PySide6.QtCore import QThread, Signal


class ProgressDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Loading Cloud File")
        self.setModal(True)
        self.progress_bar = QProgressBar(self)
        self.file_label = QLabel(self)
        layout = QVBoxLayout()
        layout.addWidget(self.file_label)
        layout.addWidget(self.progress_bar)
        self.setLayout(layout)

    def set_value(self, value):
        self.progress_bar.setValue(value)

    def set_file_name(self, file_name):
        self.file_label.setText(f"Loading: {file_name}")

    def closeEvent(self, event):
        if self.parent().progress_thread and self.parent().progress_thread.isRunning():
            event.ignore()
        else:
            event.accept()


class FileLoaderThread(QThread):
    progress = Signal(int)
    finished = Signal()

    def __init__(self, viewer, files):
        super().__init__()
        self.viewer = viewer
        self.files = files

    def run(self):
        cloud_item = self.viewer['cloud']
        for i, url in enumerate(self.files):
            file_path = url.toLocalFile()
            self.viewer.progress_dialog.set_file_name(file_path)
            cloud = cloud_item.load(file_path, append=(i > 0))
            center = np.nanmean(cloud['xyz'].astype(np.float64), axis=0)
            self.viewer.glwidget.set_cam_position(center=center)
            self.progress.emit(int((i + 1) / len(self.files) * 100))
        self.finished.emit()


class CloudViewer(q3d.Viewer):
    def __init__(self, **kwargs):
        super(CloudViewer, self).__init__(**kwargs)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Overwrite the drop event to open the cloud file.
        """
        self.progress_dialog = ProgressDialog(self)
        self.progress_dialog.show()
        files = event.mimeData().urls()
        self.progress_thread = FileLoaderThread(self, files)
        self['cloud'].load(files[0].toLocalFile(), append=False)
        self.progress_thread.progress.connect(self.file_loading_progress)
        self.progress_thread.finished.connect(self.file_loading_finished)
        self.progress_thread.start()

    def file_loading_progress(self, value):
        self.progress_dialog.set_value(value)

    def file_loading_finished(self):
        self.progress_dialog.close()

    def open_cloud_file(self, file, append=False):
        cloud_item = self['cloud']
        if cloud_item is None:
            print("Can't find clouditem.")
            return
        cloud = cloud_item.load(file, append=append)
        center = np.nanmean(cloud['xyz'].astype(np.float64), axis=0)
        self.glwidget.set_cam_position(center=center)


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--path", help="the cloud file path")
    args = parser.parse_args()
    app = q3d.QApplication(['Cloud Viewer'])
    viewer = CloudViewer(name='Cloud Viewer')
    cloud_item = q3d.CloudIOItem(size=1, alpha=0.1)
    axis_item = q3d.AxisItem(size=0.5, width=5)
    grid_item = q3d.GridItem(size=1000, spacing=20)

    viewer.add_items(
        {'cloud': cloud_item, 'grid': grid_item, 'axis': axis_item})

    if args.path:
        pcd_fn = args.path
        viewer.open_cloud_file(pcd_fn)

    viewer.show()
    app.exec()


if __name__ == '__main__':
    main()



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QPainter, QColor
from PySide6.QtWidgets import QSlider


class RangeSlider(QSlider):
    # Signal emitted when the range changes
    rangeChanged = Signal(int, int)

    def __init__(self, orientation=Qt.Horizontal,
                 parent=None, vmin=0, vmax=255):
        super().__init__(orientation, parent)
        self.setMinimum(vmin)
        self.setMaximum(vmax)
        self.lower_value = vmin
        self.upper_value = vmax
        self.active_handle = None
        self.setTickPosition(QSlider.NoTicks)  # Hide original ticks
        # Hide slider handle
        self.setStyleSheet("QSlider::handle { background: transparent; }")

    def mousePressEvent(self, event):
        """Override to handle which handle is selected."""
        pos = self.pixelPosToValue(event.pos())
        if abs(pos - self.lower_value) < abs(pos - self.upper_value):
            self.active_handle = "lower"
        else:
            self.active_handle = "upper"

    def mouseMoveEvent(self, event):
        """Override to update handle positions."""
        if event.buttons() != Qt.LeftButton:
            return

        pos = self.pixelPosToValue(event.pos())
        if self.active_handle == "lower":
            self.lower_value = max(
                self.minimum(), min(pos, self.upper_value - 1))
        elif self.active_handle == "upper":
            self.upper_value = min(
                self.maximum(), max(pos, self.lower_value + 1))
        self.rangeChanged.emit(self.lower_value, self.upper_value)
        self.update()

    def paintEvent(self, event):
        """Override to paint custom range handles."""
        painter = QPainter(self)

        # Draw the range bar
        bar_color = QColor(200, 200, 200)  # Gray bar
        highlight_color = QColor(100, 100, 255)  # Blue for selected range
        painter.setPen(Qt.NoPen)

        bar_height = 6
        bar_y = self.height() // 2 - bar_height // 2
        painter.setBrush(bar_color)
        painter.drawRect(0, bar_y, self.width(), bar_height)

        # Draw the selected range
        lower_x = int(self.valueToPixelPos(self.lower_value))
        upper_x = int(self.valueToPixelPos(self.upper_value))
        painter.setBrush(highlight_color)
        painter.drawRect(lower_x, bar_y, upper_x - lower_x, bar_height)

        # Draw the range handles
        handle_color = QColor(50, 50, 255)  # Blue handles
        painter.setBrush(handle_color)
        painter.drawEllipse(lower_x - 5, bar_y - 4, 12, 12)
        painter.drawEllipse(upper_x - 5, bar_y - 4, 12, 12)

        painter.end()

    def pixelPosToValue(self, pos):
        """Convert pixel position to slider value."""
        return self.minimum() + (self.maximum() - self.minimum()) \
            * pos.x() / self.width()

    def valueToPixelPos(self, value):
        """Convert slider value to pixel position."""
        return self.width() * (value - self.minimum()) /\
            (self.maximum() - self.minimum())



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
import numpy as np
import threading
from PySide6.QtWidgets import QLabel, QLineEdit, QDoubleSpinBox
from PySide6.QtCore import QRegularExpression
from PySide6.QtGui import QRegularExpressionValidator
from q3dviewer.utils.maths import hex_to_rgba


class LineItem(BaseItem):
    def __init__(self, width=1, color='#00ff00', line_type='LINE_STRIP'):
        super(LineItem, self).__init__()
        self.width = width
        self.buff = np.empty((0, 3), np.float32)
        self.wait_add_data = None
        self.mutex = threading.Lock()
        self.capacity = 100000
        self.valid_buff_top = 0
        self.color = color
        self.rgb = hex_to_rgba(color)
        self.line_type = GL_LINE_STRIP if line_type == 'LINE_STRIP' else GL_LINES

    def add_setting(self, layout):
        label_color = QLabel("Color:")
        layout.addWidget(label_color)
        self.color_edit = QLineEdit()
        self.color_edit.setToolTip("Hex number, i.e. #FF4500")
        self.color_edit.setText(self.color)
        self.color_edit.textChanged.connect(self._on_color)
        regex = QRegularExpression(r"^#[0-9A-Fa-f]{6}$")
        validator = QRegularExpressionValidator(regex)
        self.color_edit.setValidator(validator)
        layout.addWidget(self.color_edit)

        spinbox_width = QDoubleSpinBox()
        spinbox_width.setPrefix("Width: ")
        spinbox_width.setSingleStep(0.1)
        spinbox_width.setValue(self.width)
        spinbox_width.setRange(0.1, 10.0)
        spinbox_width.valueChanged.connect(self.set_width)
        layout.addWidget(spinbox_width)

    def _on_color(self, color):
        try:
            self.rgb = hex_to_rgba(color)
            self.color = color
        except ValueError:
            pass

    def set_color(self, color):
        self.color_edit.setText(color)

    def set_width(self, width):
        self.width = width

    def set_data(self, data, append=False):
        self.mutex.acquire()
        data = data.astype(np.float32).reshape(-1, 3)
        if (append is False):
            self.wait_add_data = data
            self.add_buff_loc = 0
        else:
            if (self.wait_add_data is None):
                self.wait_add_data = data
            else:
                self.wait_add_data = np.concatenate([self.wait_add_data, data])
            self.add_buff_loc = self.valid_buff_top
        self.mutex.release()

    def update_render_buffer(self):
        if (self.wait_add_data is None):
            return
        self.mutex.acquire()

        new_buff_top = self.add_buff_loc + self.wait_add_data.shape[0]
        if new_buff_top > self.buff.shape[0]:
            buff_capacity = self.buff.shape[0]
            while (new_buff_top > buff_capacity):
                buff_capacity += self.capacity
            self.buff = np.empty((buff_capacity, 3), np.float32)
            self.buff[self.add_buff_loc:new_buff_top] = self.wait_add_data
            glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
            glBufferData(GL_ARRAY_BUFFER, self.buff.nbytes,
                         self.buff, GL_DYNAMIC_DRAW)
            glBindBuffer(GL_ARRAY_BUFFER, 0)
        else:
            self.buff[self.add_buff_loc:new_buff_top] = self.wait_add_data
            glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
            glBufferSubData(GL_ARRAY_BUFFER, self.add_buff_loc * 12,
                            self.wait_add_data.shape[0] * 12,
                            self.wait_add_data)
        self.valid_buff_top = new_buff_top
        self.wait_add_data = None
        self.mutex.release()

    def initialize_gl(self):
        self.vbo = glGenBuffers(1)

    def paint(self):
        self.update_render_buffer()
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
        glEnableClientState(GL_VERTEX_ARRAY)
        glVertexPointer(3, GL_FLOAT, 0, None)
        glLineWidth(self.width)
        glColor4f(*self.rgb)

        glDrawArrays(self.line_type, 0, self.valid_buff_top)
        glDisableClientState(GL_VERTEX_ARRAY)

        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glDisable(GL_BLEND)



#!/usr/bin/env python3

"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
import q3dviewer as q3d
from q3dviewer.utils.cloud_io import load_gs, rotate_gaussian


class GuassianViewer(q3d.Viewer):
    def __init__(self, **kwds):
        super(GuassianViewer, self).__init__(**kwds)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        for url in event.mimeData().urls():
            file_path = url.toLocalFile()
            self.open_gs_file(file_path)

    def open_gs_file(self, file):
        gau_item = self['gaussian']
        if gau_item is None:
            print("Can't find gaussianitem")
            return

        print("Try to load %s ..." % file)
        gs = load_gs(file)
        # convert camera optical frame (b) to camera frame (c).
        Rcb = np.array([[0, -1, 0],
                        [0, 0, -1],
                        [1, 0, 0]]).T
        gs = rotate_gaussian(Rcb, gs)
        gs_data = gs.view(np.float32).reshape(gs.shape[0], -1)
        gau_item.set_data(gs_data=gs_data)


def main():
    app = q3d.QApplication(['Guassian Viewer'])
    viewer = GuassianViewer(name='Guassian Viewer')

    grid_item = q3d.GridItem(size=1000, spacing=20)
    gau_item = q3d.GaussianItem()

    viewer.add_items({'grid': grid_item, 'gaussian': gau_item})

    viewer.show()
    app.exec()


if __name__ == '__main__':
    main()



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
import numpy as np
from PySide6.QtWidgets import QDoubleSpinBox


class AxisItem(BaseItem):
    def __init__(self, size=1.0, width=2):
        super().__init__()
        self.size = size
        self.width = width
        self.T = np.eye(4, dtype=np.float32)
        self.need_update_setting = True

    def initialize_gl(self):
        # Axis vertices
        self.vertices = np.array([
            # positions
            [0.0, 0.0, 0.0], [1.0, 0.0, 0.0],  # X axis
            [0.0, 0.0, 0.0], [0.0, 1.0, 0.0],  # Y axis
            [0.0, 0.0, 0.0], [0.0, 0.0, 1.0],  # Z axis
        ], dtype=np.float32)

        # Axis colors
        self.colors = np.array([
            [1.0, 0.0, 0.0], [1.0, 0.0, 0.0],  # X axis (red)
            [0.0, 1.0, 0.0], [0.0, 1.0, 0.0],  # Y axis (green)
            [0.0, 0.0, 1.0], [0.0, 0.0, 1.0],  # Z axis (blue)
        ], dtype=np.float32)


    def add_setting(self, layout):
        spinbox_size = QDoubleSpinBox()
        spinbox_size.setPrefix("Size: ")
        spinbox_size.setSingleStep(0.1)
        spinbox_size.setValue(self.size)
        spinbox_size.setRange(0.0, 100)
        spinbox_size.valueChanged.connect(self.set_size)
        layout.addWidget(spinbox_size)

        spinbox_width = QDoubleSpinBox()
        spinbox_width.setPrefix("Width: ")
        spinbox_width.setSingleStep(0.1)
        spinbox_width.setValue(self.width)
        spinbox_width.setRange(0, 1000)
        spinbox_width.valueChanged.connect(self.set_width)
        layout.addWidget(spinbox_width)

    def set_size(self, size):
        self.size = size

    def set_width(self, width):
        self.width = width
        
    def set_transform(self, transform):
        """
        Set the transformation matrix for the axis item.
        """
        self.T = transform
        self.need_update_setting = True

    def paint(self):
        glLineWidth(self.width)

        glPushMatrix()
        glMultMatrixf(self.T.T)

        glEnableClientState(GL_VERTEX_ARRAY)
        glVertexPointer(3, GL_FLOAT, 0, self.vertices * self.size)

        glEnableClientState(GL_COLOR_ARRAY)
        glColorPointer(3, GL_FLOAT, 0, self.colors)

        glDrawArrays(GL_LINES, 0, len(self.vertices))

        glDisableClientState(GL_VERTEX_ARRAY)
        glDisableClientState(GL_COLOR_ARRAY)
        glPopMatrix()

        glLineWidth(1)



#!/usr/bin/env python3

import numpy as np
import q3dviewer as q3d
from PySide6.QtCore import QTimer

i = 0.0

def update(viewer):
    global i
    i += 0.05
    new = np.array([np.sin(i) * i, np.cos(i) * i, i])
    viewer['traj'].set_data(new, append=True)
    viewer.update()
    return i


def main():
    # add a qt application
    app = q3d.QApplication([])

    # create opengl items
    axis_item = q3d.AxisItem(size=0.5, width=5)
    grid_item = q3d.GridItem(size=10, spacing=1)
    traj_item = q3d.LineItem(width=2)

    # create viewer
    viewer = q3d.Viewer(name='example')

    # add all items to viewer
    viewer.add_items({
        'grid': grid_item,
        'axis': axis_item,
        'traj': traj_item})

    # dynamic update by timer
    timer = QTimer()
    timer.timeout.connect(lambda: update(viewer))
    timer.start(100)

    # show viewer and start application
    viewer.show()
    app.exec()


if __name__ == '__main__':
    main()



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from PySide6 import QtCore, QtGui
from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
from q3dviewer.utils.maths import hex_to_rgba


class Text2DItem(BaseItem):
    """Draws text over opengl 3D."""

    def __init__(self, **kwds):
        """All keyword arguments are passed to set_data()"""
        BaseItem.__init__(self)
        self.pos = (20, 50)
        self.color = '#ffffff'
        self.rgb = hex_to_rgba(self.color)
        self.text = ''
        self.font = QtGui.QFont('Helvetica', 16)
        self.set_data(**kwds)

    def set_data(self, **kwds):
        args = ['pos', 'color', 'text', 'size', 'font']
        for k in kwds.keys():
            if k not in args:
                raise ValueError('Invalid keyword argument: %s\
                    (allowed arguments are %s)' % (k, str(args)))
        for arg in args:
            if arg in kwds:
                value = kwds[arg]
                if arg == 'pos':
                    self.pos = value
                elif arg == 'color':
                    self.set_color(value)
                elif arg == 'font':
                    if isinstance(value, QtGui.QFont) is False:
                        raise TypeError('"font" must be QFont.')
                elif arg == 'size':
                    self.font.setPointSize(value)
                setattr(self, arg, value)

    def set_color(self, color):
        try:
            self.rgb = hex_to_rgba(color)
            self.color = color
        except ValueError:
            pass

    def paint(self):
        if len(self.text) < 1:
            return

        text_pos = QtCore.QPointF(*self.pos)
        painter = QtGui.QPainter(self.glwidget())
        painter.setPen(QtGui.QColor(*[int(c * 255) for c in self.rgb[:3]]))
        painter.setFont(self.font)
        painter.setRenderHints(QtGui.QPainter.RenderHint.Antialiasing |
                               QtGui.QPainter.RenderHint.TextAntialiasing)
        painter.drawText(text_pos, self.text)
        painter.end()



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

"""
Math proofs and implementations: 
https://github.com/scomup/MathematicalRobotics.git
"""


import numpy as np


_epsilon_ = 1e-5


def skew(vector):
    return np.array([[0, -vector[2], vector[1]],
                     [vector[2], 0, -vector[0]],
                     [-vector[1], vector[0], 0]])


def expSO3(omega):
    """
    Exponential map of SO3
    The proof is shown in 3d_rotation_group.md (10)
    """
    theta2 = omega.dot(omega)
    theta = np.sqrt(theta2)
    nearZero = theta2 <= _epsilon_
    W = skew(omega)
    if (nearZero):
        return np.eye(3) + W
    else:
        K = W/theta
        KK = K.dot(K)
        sin_theta = np.sin(theta)
        one_minus_cos = 1 - np.cos(theta)
        R = np.eye(3) + sin_theta * K + one_minus_cos * KK  # rotation.md (10)
        return R


def logSO3(R):
    """
    Logarithm map of SO3
    The proof is shown in rotation.md (14)
    """
    R11, R12, R13 = R[0, :]
    R21, R22, R23 = R[1, :]
    R31, R32, R33 = R[2, :]
    tr = np.trace(R)
    omega = np.zeros(3)
    v = np.array([R32 - R23, R13 - R31, R21 - R12])
    # when trace == -1, i.e., the theta is approx to +-pi, +-3pi, +-5pi, etc.
    # we do something special
    if (tr + 1.0 < 1e-3):
        if (R33 > R22 and R33 > R11):
            # R33 is largest
            # sin(theta) approx to sgn_w*pi-theta, cos(theta) approx to -1
            W = R21 - R12  # 2*r3*sin(theta) = 2*r3*(sgn_w*pi-theta)
            Q1 = R31 + R13          # 4 * r1*r3
            Q2 = R23 + R32          # 4 * r2*r3
            Q3 = 2.0 + 2.0 * R33    # 4 * r3*r3
            r = np.sqrt(Q3)         # 2 * r3
            one_over_r = 1 / r      # 1 / (2*r3)
            norm = np.sqrt(Q1*Q1 + Q2*Q2 + Q3*Q3 + W*W)  # 4*r3
            sgn_w = np.sign(W)      # get the sgn of theta
            mag = np.pi - (2 * sgn_w * W) / norm   # theta*sgn_w
            scale = 0.5 * mag * one_over_r  # theta * sgn_w / (4*r3)
            # omega = theta * [4*r1*r3, 4*r2*r3, 4*r3*r3]/ (4*r3)
            omega = sgn_w * scale * np.array([Q1, Q2, Q3])
        elif (R22 > R11):
            # R22 is the largest
            W = R13 - R31  # 2*r2*sin(theta) = 2*r2*(sgn_w*pi-theta)
            Q1 = R12 + R21        # 4 * r2*r1
            Q2 = 2.0 + 2.0 * R22  # 4 * r2*r2
            Q3 = R23 + R32        # 4 * r2*r3
            r = np.sqrt(Q2)
            one_over_r = 1 / r
            norm = np.sqrt(Q1*Q1 + Q2*Q2 + Q3*Q3 + W*W)
            sgn_w = np.sign(W)
            mag = np.pi - (2 * sgn_w * W) / norm
            scale = 0.5 * one_over_r * mag
            omega = sgn_w * scale * np.array([Q1, Q2, Q3])
        else:
            # R11 is the largest
            W = R32 - R23  # 2*r1*sin(theta) = 2*r1*(sgn_w*pi-theta)
            Q1 = 2.0 + 2.0 * R11 # 4 * r1*r1
            Q2 = R12 + R21       # 4 * r1*r2
            Q3 = R31 + R13       # 4 * r1*r3
            r = np.sqrt(Q1)
            one_over_r = 1 / r
            norm = np.sqrt(Q1*Q1 + Q2*Q2 + Q3*Q3 + W*W)
            sgn_w = np.sign(W)
            mag = np.pi - (2 * sgn_w * W) / norm
            scale = 0.5 * one_over_r * mag
            omega = sgn_w * scale * np.array([Q1, Q2, Q3])
    else:
        magnitude = 0
        tr_3 = tr - 3.0
        if (tr_3 < -1e-6):
            # this is the normal case -1 < trace < 3
            theta = np.arccos((tr - 1.0) / 2.0)
            magnitude = theta / (2.0 * np.sin(theta))
        else:
            # when theta near 0, +-2pi, +-4pi, etc. (trace near 3.0)
            # use Taylor expansion: theta \approx 1/2-(t-3)/12 + O((t-3)^2)
            # see https://github.com/borglab/gtsam/issues/746 for details
            magnitude = 0.5 - tr_3 / 12.0 + tr_3*tr_3/60.0
        omega = magnitude * np.array([R32 - R23, R13 - R31, R21 - R12])
    return omega


def interpolate_pose(T1, T2, v_max, omega_max, dt=0.02):
    R1, t1 = makeRt(T1)
    R2, t2 = makeRt(T2)
    
    # Get transfrom time based on linear velocity
    d = np.linalg.norm(t2 - t1)
    t_lin = d / v_max
    
    # Get transform time based on angular velocity
    omega = logSO3(R2 @ R1.T)
    theta = np.linalg.norm(omega)
    t_ang = theta / omega_max
    
    # Get total time based on the linear and angular time.
    t_total = max(t_lin, t_ang)
    num_steps = int(np.ceil(t_total / dt))
    
    # Generate interpolated transforms
    interpolated_Ts = []
    for i in range(num_steps):
        s = i / num_steps
        t_interp = (1 - s) * t1 + s * t2
        # Interpolate rotation using SO3.
        R_interp = expSO3(s * omega) @ R1
        T_interp = makeT(R_interp, t_interp)
        interpolated_Ts.append(T_interp)
    
    return interpolated_Ts


def frustum(left, right, bottom, top, near, far):
    # see https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/glFrustum.xml
    if near <= 0 or far <= 0 or near >= far or left == right or bottom == top:
        print("Invalid frustum parameters.")
        return None
    matrix = np.zeros((4, 4), dtype=np.float32)
    matrix[0, 0] = 2.0 * near / (right - left)
    matrix[0, 2] = (right + left) / (right - left)
    matrix[1, 1] = 2.0 * near / (top - bottom)
    matrix[1, 2] = (top + bottom) / (top - bottom)
    matrix[2, 2] = -(far + near) / (far - near)
    matrix[2, 3] = -2.0 * far * near / (far - near)
    matrix[3, 2] = -1.0
    return matrix


def rainbow(scalars, scalar_min=0, scalar_max=255):
    range = scalar_max - scalar_min
    values = 1.0 - (scalars - scalar_min) / range
    # values = (scalars - scalar_min) / range  # using inverted color
    colors = np.zeros([scalars.shape[0], 3], dtype=np.float32)
    values = np.clip(values, 0, 1)

    h = values * 5.0 + 1.0
    i = np.floor(h).astype(int)
    f = h - i
    f[np.logical_not(i % 2)] = 1 - f[np.logical_not(i % 2)]
    n = 1 - f

    # idx = i <= 1
    colors[i <= 1, 0] = n[i <= 1] * 255
    colors[i <= 1, 1] = 0
    colors[i <= 1, 2] = 255

    colors[i == 2, 0] = 0
    colors[i == 2, 1] = n[i == 2] * 255
    colors[i == 2, 2] = 255

    colors[i == 3, 0] = 0
    colors[i == 3, 1] = 255
    colors[i == 3, 2] = n[i == 3] * 255

    colors[i == 4, 0] = n[i == 4] * 255
    colors[i == 4, 1] = 255
    colors[i == 4, 2] = 0

    colors[i >= 5, 0] = 255
    colors[i >= 5, 1] = n[i >= 5] * 255
    colors[i >= 5, 2] = 0
    return colors


def euler_to_matrix(rpy):
    roll, pitch, yaw = rpy
    Rx = np.array([[1, 0, 0],
                   [0, np.cos(roll), -np.sin(roll)],
                   [0, np.sin(roll), np.cos(roll)]])
    Ry = np.array([[np.cos(pitch), 0, np.sin(pitch)],
                   [0, 1, 0],
                   [-np.sin(pitch), 0, np.cos(pitch)]])
    Rz = np.array([[np.cos(yaw), -np.sin(yaw), 0],
                   [np.sin(yaw), np.cos(yaw), 0],
                   [0, 0, 1]])
    R = Rz @ Ry @ Rx
    return R


def matrix_to_euler(R):
    sy = np.sqrt(R[0, 0]**2 + R[1, 0]**2)
    singular = sy < 1e-6  # Check for gimbal lock
    if not singular:
        roll = np.arctan2(R[2, 1], R[2, 2])   # X-axis rotation
        pitch = np.arctan2(-R[2, 0], sy)      # Y-axis rotation
        yaw = np.arctan2(R[1, 0], R[0, 0])    # Z-axis rotation
    else:
        # Gimbal lock case
        roll = np.arctan2(-R[1, 2], R[1, 1])
        pitch = np.arctan2(-R[2, 0], sy)
        yaw = 0  # Arbitrarily set yaw to 0

    return np.array([roll, pitch, yaw])


def matrix_to_quaternion(matrix):
    trace = matrix[0, 0] + matrix[1, 1] + matrix[2, 2]
    if trace > 0:
        s = 0.5 / np.sqrt(trace + 1.0)
        w = 0.25 / s
        x = (matrix[2, 1] - matrix[1, 2]) * s
        y = (matrix[0, 2] - matrix[2, 0]) * s
        z = (matrix[1, 0] - matrix[0, 1]) * s
    else:
        if matrix[0, 0] > matrix[1, 1] and matrix[0, 0] > matrix[2, 2]:
            s = 2.0 * np.sqrt(1.0 + matrix[0, 0] - matrix[1, 1] - matrix[2, 2])
            w = (matrix[2, 1] - matrix[1, 2]) / s
            x = 0.25 * s
            y = (matrix[0, 1] + matrix[1, 0]) / s
            z = (matrix[0, 2] + matrix[2, 0]) / s
        elif matrix[1, 1] > matrix[2, 2]:
            s = 2.0 * np.sqrt(1.0 + matrix[1, 1] - matrix[0, 0] - matrix[2, 2])
            w = (matrix[0, 2] - matrix[2, 0]) / s
            x = (matrix[0, 1] + matrix[1, 0]) / s
            y = 0.25 * s
            z = (matrix[1, 2] + matrix[2, 1]) / s
        else:
            s = 2.0 * np.sqrt(1.0 + matrix[2, 2] - matrix[0, 0] - matrix[1, 1])
            w = (matrix[1, 0] - matrix[0, 1]) / s
            x = (matrix[0, 2] + matrix[2, 0]) / s
            y = (matrix[1, 2] + matrix[2, 1]) / s
            z = 0.25 * s
    return np.array([x, y, z, w])


def quaternion_to_matrix(quaternion):
    x, y, z, w = quaternion
    q = np.array(quaternion[:4], dtype=np.float64, copy=True)
    n = np.linalg.norm(q)
    if np.any(n == 0.0):
        raise ZeroDivisionError("bad quaternion input")
    else:
        m = np.empty((3, 3))
        m[0, 0] = 1.0 - 2*(y**2 + z**2)/n
        m[0, 1] = 2*(x*y - z*w)/n
        m[0, 2] = 2*(x*z + y*w)/n
        m[1, 0] = 2*(x*y + z*w)/n
        m[1, 1] = 1.0 - 2*(x**2 + z**2)/n
        m[1, 2] = 2*(y*z - x*w)/n
        m[2, 0] = 2*(x*z - y*w)/n
        m[2, 1] = 2*(y*z + x*w)/n
        m[2, 2] = 1.0 - 2*(x**2 + y**2)/n
        return m


def make_transform(pose, rotation):
    transform = np.eye(4)
    transform[0:3, 0:3] = quaternion_to_matrix(rotation)
    transform[0:3, 3] = pose
    return transform

def makeT(R, t):
    T = np.eye(4)
    T[0:3, 0:3] = R
    T[0:3, 3] = t
    return T

def makeRt(T):
    R = T[0:3, 0:3]
    t = T[0:3, 3]
    return R, t

def hex_to_rgba(hex_color):
    if not hex_color.startswith("#"):
        print("Invalid hex color string.")
        return (1.0, 1.0, 1.0, 1.0)
    if len(hex_color) == 7:
        color_flat = int(hex_color[1:], 16)
        red = (color_flat >> 16) & 0xFF
        green = (color_flat >> 8) & 0xFF
        blue = color_flat & 0xFF
        return (red / 255.0, green / 255.0, blue / 255.0, 1.0)
    elif len(hex_color) == 9:
        color_flat = int(hex_color[1:], 16)
        red = (color_flat >> 24) & 0xFF
        green = (color_flat >> 16) & 0xFF
        blue = (color_flat >> 8) & 0xFF
        alpha = color_flat & 0xFF
        return (red / 255.0, green / 255.0, blue / 255.0, alpha / 255.0)
    else:
        print("Invalid hex color string.")
        return (1.0, 1.0, 1.0, 1.0)

# euler = np.array([1, 0.1, 0.1])
# euler_angles = matrix_to_euler(euler_to_matrix(euler))
# print("Euler Angles:", euler_angles)



from q3dviewer.custom_items.axis_item import *
from q3dviewer.custom_items.cloud_item import *
from q3dviewer.custom_items.cloud_io_item import *
from q3dviewer.custom_items.gaussian_item import *
from q3dviewer.custom_items.frame_item import *
from q3dviewer.custom_items.grid_item import *
from q3dviewer.custom_items.text_item import *
from q3dviewer.custom_items.image_item import *
from q3dviewer.custom_items.line_item import *



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.base_item import BaseItem
from OpenGL.GL import *
import numpy as np
from OpenGL.GL import shaders
from q3dviewer.utils import *


# Vertex and Fragment shader source code
vertex_shader_source = """
#version 330 core
layout(location = 0) in vec3 position;
layout(location = 1) in vec2 texCoord;

out vec2 TexCoord;

uniform mat4 view_matrix;
uniform mat4 project_matrix;
uniform mat4 model_matrix;

void main()
{
    gl_Position = project_matrix * view_matrix * model_matrix * vec4(position, 1.0);
    TexCoord = texCoord;
}
"""

fragment_shader_source = """
#version 330 core
in vec2 TexCoord;
out vec4 color;
uniform sampler2D ourTexture;
void main()
{
    color = texture(ourTexture, TexCoord);
}
"""


class FrameItem(BaseItem):
    def __init__(self, T=np.eye(4), size=(1, 0.8), width=3, img=None, color='#0000FF'):
        BaseItem.__init__(self)
        self.w, self.h = size
        self.width = width
        self.T = T
        self.img = img
        self.texture = None
        self.need_updating = False
        self.set_color(color)

    def initialize_gl(self):
        # Rectangle vertices and texture coordinates
        hw = self.w / 2
        hh = self.h / 2
        self.vertices = np.array([
            # positions          # texture coords
            [-hw,  hh,  0.0,  0.0, 0.0],  # bottom-left
            [ hw,  hh,  0.0,  1.0, 0.0],  # bottom-right
            [ hw, -hh,  0.0,  1.0, 1.0],  # top-right
            [-hw, -hh,  0.0,  0.0, 1.0],  # top-left
            [ 0.0,  0.0, hh * 0.66, 0.0, 0.0],  # center -Z is the front.
        ], dtype=np.float32)

        self.indices = np.array([
            0, 1, 2,  # first triangle
            2, 3, 0   # second triangle
        ], dtype=np.uint32)

        self.vao = glGenVertexArrays(1)
        self.vbo = glGenBuffers(1)
        self.ebo = glGenBuffers(1)

        glBindVertexArray(self.vao)

        glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
        glBufferData(GL_ARRAY_BUFFER, self.vertices.itemsize *
                     5 * 4, self.vertices, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.ebo)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                     self.indices.nbytes, self.indices, GL_STATIC_DRAW)

        # Vertex positions
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE,
                              20, ctypes.c_void_p(0))
        glEnableVertexAttribArray(0)
        # Texture coordinates
        glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE,
                              20, ctypes.c_void_p(12))
        glEnableVertexAttribArray(1)
        project_matrix = self.glwidget().get_projection_matrix()
        # Compile shaders and create shader program
        self.program = shaders.compileProgram(
            shaders.compileShader(vertex_shader_source, GL_VERTEX_SHADER),
            shaders.compileShader(fragment_shader_source, GL_FRAGMENT_SHADER),
        )
        glUseProgram(self.program)
        set_uniform(self.program, np.eye(4), 'model_matrix')
        set_uniform(self.program, project_matrix, 'project_matrix')
        glUseProgram(0)
        self.texture = glGenTextures(1)
        self.set_data(img=self.img)
        
        # Define line vertices
        self.line_vertices = np.array([
            self.vertices[0, :3], self.vertices[1, :3],
            self.vertices[1, :3], self.vertices[2, :3],
            self.vertices[2, :3], self.vertices[3, :3],
            self.vertices[3, :3], self.vertices[0, :3],
            self.vertices[4, :3], self.vertices[0, :3],
            self.vertices[4, :3], self.vertices[1, :3],
            self.vertices[4, :3], self.vertices[2, :3],
            self.vertices[4, :3], self.vertices[3, :3]
        ], dtype=np.float32)

        self.line_vbo = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, self.line_vbo)
        glBufferData(GL_ARRAY_BUFFER, self.line_vertices.nbytes, self.line_vertices, GL_STATIC_DRAW)
        glBindBuffer(GL_ARRAY_BUFFER, 0)

        glBindVertexArray(0)

    def set_transform(self, Twc, is_opencv_coord=False):
        if is_opencv_coord:
            # convert the opencv camera coordinate to the opengl camera coordinate
            M_conv = np.array([
                [1, 0, 0, 0],
                [0, -1, 0, 0],
                [0, 0, -1, 0],
                [0, 0, 0, 1]
            ])
            Twc = Twc @ M_conv
        self.Twc = Twc

    def set_data(self, img=None, transform=None, is_opencv_coord=False):
        if transform is not None:
            self.set_transform(transform, is_opencv_coord)
        self.img = img
        self.need_updating = True

    def update_img_buffer(self):
        if self.need_updating:
            self.texture = glGenTextures(1)
            glBindTexture(GL_TEXTURE_2D, self.texture)
            if self.img.ndim == 2:
                # Convert grayscale to RGBA
                self.img = np.stack((self.img,) * 3 + (np.ones_like(self.img) * 255,), axis=-1)
            elif self.img.shape[2] == 3:
                # Add an alpha channel
                alpha_channel = np.ones((self.img.shape[0], self.img.shape[1], 1), dtype=np.uint8) * 255
                self.img = np.concatenate((self.img, alpha_channel), axis=2)
            img = np.ascontiguousarray(self.img, dtype=np.uint8)
            # Load image
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img.shape[1],
                         img.shape[0], 0, GL_RGBA, GL_UNSIGNED_BYTE, img)
            glGenerateMipmap(GL_TEXTURE_2D)
            glBindTexture(GL_TEXTURE_2D, 0)
            self.need_updating = False
        
    def set_color(self, color):
        if isinstance(color, str):
            self.rgba = hex_to_rgba(color)
        elif isinstance(color, list):
            self.rgba = color
        elif isinstance(color, tuple):
            self.rgba = list(color)
        else:
            raise ValueError("Invalid color format")

    def set_line_width(self, width):
        self.width = width

    def paint(self):
        self.view_matrix = self.glwidget().view_matrix
        project_matrix = self.glwidget().projection_matrix

        glEnable(GL_DEPTH_TEST)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        if self.img is not None:
            self.update_img_buffer()
            glUseProgram(self.program)
            set_uniform(self.program, self.view_matrix, 'view_matrix')
            set_uniform(self.program, project_matrix, 'project_matrix')
            set_uniform(self.program, self.T, 'model_matrix')
            glBindVertexArray(self.vao)
            glBindVertexArray(0)
            glUseProgram(0)

        # if self.texture is not None:
        #     glBindTexture(GL_TEXTURE_2D, self.texture)
        #     glMultMatrixf(self.T.T)  # Apply the transformation matrix
        #     glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, None)
        #     glBindTexture(GL_TEXTURE_2D, 0)
        
        glLineWidth(self.width)
        # set color to bule
        glColor4f(*self.rgba)
        glBindBuffer(GL_ARRAY_BUFFER, self.line_vbo)
        glEnableClientState(GL_VERTEX_ARRAY)
        glMultMatrixf(self.T.T)
        glVertexPointer(3, GL_FLOAT, 0, None)
        glDrawArrays(GL_LINES, 0, len(self.line_vertices))
        glDisableClientState(GL_VERTEX_ARRAY)
        glBindBuffer(GL_ARRAY_BUFFER, 0)

        glDisable(GL_DEPTH_TEST)
        glDisable(GL_BLEND)

